import logging
import glob
import re
import os
import shutil
import filecmp
import shlex
from xdg import DesktopEntry
from desktop_entry_lib import DesktopEntry as JdDesktopEntry

import dataclasses
from ..models.Settings import Settings
from ..models.AppListElement import AppListElement, InstalledStatus
from ..lib.constants import TMP_DIR
from ..lib import terminal
from ..lib.async_utils import idle
from ..lib.ini_config import Config
from ..lib.utils import get_giofile_content_type, gio_copy, get_file_hash, \
    remove_special_chars, get_random_string, get_osinfo, extract_terminal_arguments, show_message_dialog, gnu_naturalsize
from ..models.Models import AppUpdateElement, InternalError, DownloadInterruptedException
from typing import Optional, List, TypedDict
from gi.repository import GLib, Gtk, Gdk, Gio
from enum import Enum


class ExtractedAppImage():
    extraction_folder: str
    desktop_entry: Optional[DesktopEntry.DesktopEntry]
    appimage_file: Gio.File
    desktop_file: Optional[Gio.File]
    icon_file: Optional[Gio.File]
    md5: str

class AppImageUpdateLogic(Enum):
    REPLACE = 'REPLACE'
    KEEP = 'KEEP'

class AppImageArchitecture:
    UNKNOWN = 'UNKNOWN'
    X86_64 = 'x86_64'
    ARM_64 = 'aarch64'

@dataclasses.dataclass
class AppImageListElement():
    name: str
    description: str
    provider: str
    installed_status: InstalledStatus
    file_path: str
    trusted: bool = False
    is_updatable_from_url = False
    env_variables: List[str] = dataclasses.field(default_factory=lambda: [])
    exec_arguments: str = ''
    desktop_entry: Optional[DesktopEntry.DesktopEntry] = None
    update_logic: Optional[AppImageUpdateLogic] = None
    architecture: Optional[AppImageArchitecture] = None
    updating_from: Optional[any] = None # AppImageListElement
    version: Optional[str] = None
    extracted: Optional[ExtractedAppImage] = None
    local_file: Optional[bool] = None
    external_folder: bool = False
    desktop_file_path: Optional[str] = None
    size: int = 0

    def set_installed_status(self, installed_status: InstalledStatus):
        self.installed_status = installed_status

    def set_trusted(self):
        logging.debug('Chmod file ' + self.file_path)
        os.chmod(self.file_path, 0o755)
        self.trusted = True

    def get_config(self):
        return Config.get_app_config(self)

class AppImageProvider():
    # https://cgit.freedesktop.org/xdg/shared-mime-info/tree/data/freedesktop.org.xml.in#n2114
    supported_mimes = ['application/x-iso9660-appimage', 'application/vnd.appimage', 'application/x-appimage',
                    'application/vnd.efi.iso']

    def __init__(self):
        self.name = 'AppImage'
        self.v2_detector_string = 'AppImages require FUSE to run.'
        self.icon = "/it/mijorus/gearlever/assets/App-image-logo.png"
        self.desktop_exec_codes = ["%f", "%F",  "%u",  "%U",  "%i",  "%c", "%k"]
        logging.info(f'Activating {self.name} provider')

        self.general_messages = []
        self.update_messages = []

        self.extraction_folder = os.path.join(TMP_DIR, 'appimages')
        self.user_desktop_files_path = os.path.join(GLib.get_home_dir(), '.local', 'share', 'applications')
        self.user_local_share_path = os.path.join(GLib.get_home_dir(), '.local', 'share')
    desk_entry_section_regex = re.compile(r'\[Desktop Entry\][\s\S]*?(?=\n\[)', flags=re.MULTILINE)

    def list_installed(self) -> list[AppImageListElement]:
        default_folder_path = self._get_appimages_default_destination_path()
        manage_from_outside = Settings.settings.get_boolean('manage-files-outside-default-folder')
        output = []

        if not os.path.exists(self.user_desktop_files_path):
            return output

        for file_name in os.listdir(self.user_desktop_files_path):
            gfile = Gio.File.new_for_path(
                os.path.join(self.user_desktop_files_path, file_name))

            try:
                if os.path.isfile(gfile.get_path()) and get_giofile_content_type(gfile) == 'application/x-desktop':
                    entry = DesktopEntry.DesktopEntry(filename=gfile.get_path())
                    exec_location = entry.getTryExec()
                    exec_command_data = extract_terminal_arguments(entry.getExec())
                    exec_arguments = ' '.join(exec_command_data['arguments'])

                    if os.path.isfile(exec_location):
                        exec_gfile = Gio.File.new_for_path(exec_location)
                        exec_in_defalut_folder = os.path.isfile(
                                os.path.join(default_folder_path, exec_gfile.get_basename()))
                        exec_in_folder = True if manage_from_outside else exec_in_defalut_folder
                        app_version = self._get_app_version(None, desktop_entry=entry, return_hash=False)
                        file_size = os.stat(exec_location).st_size

                        if exec_in_folder and self.can_install_file(exec_gfile):
                            list_element = AppImageListElement(
                                name=entry.getName(),
                                desktop_file_path=gfile.get_path(),
                                description=entry.getComment(),
                                version=(app_version or None),
                                installed_status=InstalledStatus.INSTALLED,
                                file_path=exec_location,
                                provider=self.name,
                                desktop_entry=entry,
                                trusted=True,
                                external_folder=(not exec_in_defalut_folder),
                                exec_arguments=exec_arguments,
                                env_variables=exec_command_data['env_vars'],
                                size=file_size,
                            )

                            list_element.architecture = None

                            output.append(list_element)
                        else:
                            logging.debug(f'{gfile.get_path()} skipped because {exec_location} is not a supported file type')
                    else:
                        logging.debug(f'{gfile.get_path()} skipped because {exec_location} does not exists on the filesystem')

            except Exception as e:
                logging.warn(e)

        return output

    def is_installed(self, el: AppImageListElement) -> bool:
        if el.file_path and os.path.exists(self._get_appimages_default_destination_path()):
            for file_name in os.listdir(self._get_appimages_default_destination_path()):
                installed_gfile = Gio.File.new_for_path(self._get_appimages_default_destination_path() + '/' + file_name)
                loaded_gfile = Gio.File.new_for_path(el.file_path)

                if get_giofile_content_type(installed_gfile) in self.supported_mimes:
                    if filecmp.cmp(installed_gfile.get_path(), loaded_gfile.get_path(), shallow=False):
                        el.file_path = installed_gfile.get_path()
                        return True

        return False

    def get_icon(self, el: AppImageListElement) -> Gtk.Image:
        icon_path = None

        if el.desktop_entry:
            icon_path = el.desktop_entry.getIcon()

        if icon_path and os.path.isfile(icon_path):
            return Gtk.Image.new_from_file(icon_path)
        else:
            icon_theme = Gtk.IconTheme.get_for_display(Gdk.Display.get_default())

            if el.desktop_entry and icon_theme.has_icon(el.desktop_entry.getIcon()):
                return Gtk.Image.new_from_icon_name(el.desktop_entry.getIcon())

            extracted = self._load_appimage_metadata(el)

            if extracted.icon_file and os.path.exists(extracted.icon_file.get_path()):
                return Gtk.Image.new_from_file(extracted.icon_file.get_path())

        return Gtk.Image(icon_name='gl-application-x-executable-symbolic')

    def get_description(self, el: AppImageListElement) -> str:
        if el.desktop_entry:
            return el.desktop_entry.getComment()

        return ''

    def refresh_data(self, el: AppImageListElement):
        if el.desktop_entry:
            el.name = el.desktop_entry.getName()

        extracted = self._load_appimage_metadata(el)
        if extracted.desktop_entry:
            el.name = extracted.desktop_entry.getName()
            el.version = el.desktop_entry.get('X-AppImage-Version')

    def refresh_arch(self, el: AppImageListElement):
        el.architecture = self.get_elf_arch(el)

    def uninstall(self, el: AppImageListElement, force_delete=False, remove_configuration=True):
        logging.info(f'Removing {el.file_path}')

        gf = Gio.File.new_for_path(el.file_path)

        if force_delete:
            os.remove(el.file_path)
        else:
            try:
                logging.info(f'Trashing {el.file_path}')
                gf.trash(None)
            except Exception as e:
                logging.warn(f'Trashing {el.file_path} failed! Removing it instead...')
                logging.warn(e)
                os.remove(el.file_path)

        if el.desktop_entry:
            logging.info(f'Removing {el.desktop_entry.getFileName()}')
            os.remove(el.desktop_entry.getFileName())

        icon = el.desktop_entry.getIcon()
        if '/' in icon and os.path.isfile(icon):
            os.remove(icon)

        if remove_configuration:
            Config.delete_app_update_config(el)

        el.set_installed_status(InstalledStatus.NOT_INSTALLED)
        Config.delete_app_config(el)
        Config.delete_app_update_config(el)

    def run(self, el: AppImageListElement):
        if el.trusted:
            if el.installed_status is InstalledStatus.INSTALLED:
                gtk_launch = False

                try:
                    terminal.host_sh(['which', 'gtk-launch'])
                    gtk_launch = True
                except Exception as e:
                    logging.warning('gtk-launch is missing, falling back to executable launch')

                if gtk_launch and el.desktop_file_path:
                    desktop_file_name = os.path.basename(el.desktop_file_path)
                    terminal.host_threaded_sh(['gtk-launch', desktop_file_name], callback=self._check_launch_output, return_stderr=True)
                else:
                    self._run_from_desktopentry(el)
            else:
                self._run_filepath(el)

    def can_install_file(self, file: Gio.File) -> bool:
        return get_giofile_content_type(file) in self.supported_mimes

    def is_updatable(self, el: AppImageListElement) -> bool:
        for item in self.list_installed():
            if item.name == el.name:
                return True

        return False

    def install_file(self, el: AppImageListElement):
        logging.info('Installing appimage: ' + el.file_path)
        el.installed_status = InstalledStatus.INSTALLING
        extracted_appimage: Optional[ExtractedAppImage] = None
        appimages_destination_path = self._get_appimages_default_destination_path()

        try:
            extracted_appimage = self._load_appimage_metadata(el)
            version = self._get_app_version(extracted_appimage, return_hash=False)
            dest_file_info = extracted_appimage.appimage_file.query_info('*', Gio.FileQueryInfoFlags.NOFOLLOW_SYMLINKS)

            # Move .appimage to its default location
            if not os.path.exists(f'{appimages_destination_path}'):
                os.mkdir(f'{appimages_destination_path}')

            # how the appimage will be called
            appimage_filename = ''
            prefixed_filename = ''
            if el.update_logic == AppImageUpdateLogic.REPLACE and el.updating_from is not None:
                appimage_filename = os.path.basename(el.updating_from.file_path)
                desktop_file_path = os.path.basename(el.updating_from.desktop_file_path)
                prefixed_filename = os.path.splitext(desktop_file_path)[0]
            else:
                dest_file_info_name = os.path.splitext(dest_file_info.get_name())[0]
                appimage_filename = f'gearlever_{dest_file_info_name}'

                if extracted_appimage.desktop_entry:
                    appimage_filename = extracted_appimage.desktop_entry.getName()
                    appimage_filename = appimage_filename.lower().replace(' ', '_')

                append_file_ext = True
                gsettings = Settings.settings

                if extracted_appimage.desktop_entry and \
                    gsettings.get_boolean('exec-as-name-for-terminal-apps') and \
                        extracted_appimage.desktop_entry.getTerminal():

                    append_file_ext = False
                    if appimage_filename == 'AppDir':
                        appimage_filename = extracted_appimage.desktop_entry.getName()

                if append_file_ext:
                    appimage_filename = f'{appimage_filename}.appimage'

                app_name_without_ext = appimage_filename
                appimage_filename = remove_special_chars(appimage_filename).lower()

                i = 0
                files_in_dest_dir = os.listdir(self._get_appimages_default_destination_path())

                # if there is already an app with the same name,
                # we try not to overwrite
                while appimage_filename in files_in_dest_dir:
                    if i == 0 and version:
                        appimage_filename = app_name_without_ext + '_' + version.replace('.', '_')
                    else:
                        appimage_filename = app_name_without_ext + f'_{i}'

                    if append_file_ext:
                        appimage_filename = appimage_filename + '.appimage'

                    i += 1

                prefixed_filename = os.path.splitext(appimage_filename)[0]

            dest_appimage_file = Gio.File.new_for_path(
                os.path.join(appimages_destination_path, appimage_filename))

            if not gio_copy(extracted_appimage.appimage_file, dest_appimage_file):
                raise InternalError('Error while moving appimage file to the destination folder')

            logging.debug(f'file copied to {appimages_destination_path}')

            el.file_path = dest_appimage_file.get_path()
            el.set_trusted()

            # copy the icon file
            icon_file = None
            dest_appimage_icon_file = None
            if extracted_appimage.desktop_entry:
                icon_file = extracted_appimage.icon_file

            if icon_file and os.path.exists(icon_file.get_path()):
                icons_folder = os.path.join(appimages_destination_path, '.icons')
                if not os.path.exists(icons_folder):
                    os.mkdir(icons_folder)

                i, icon_file_ext = os.path.splitext(icon_file.get_path())
                dest_appimage_icon_file_path = os.path.join(appimages_destination_path, '.icons', prefixed_filename)
                dest_appimage_icon_file = Gio.File.new_for_path(
                    f'{dest_appimage_icon_file_path}{icon_file_ext}')

                gio_copy(icon_file, dest_appimage_icon_file)

            # Move .desktop file to its default location
            if not os.path.exists(self.user_desktop_files_path):
                os.makedirs(self.user_desktop_files_path)

            dest_desktop_file_path = f'{os.path.join(self.user_desktop_files_path, prefixed_filename)}.desktop'
            dest_desktop_file_path = dest_desktop_file_path.replace(' ', '_')

            # Get default exec arguments
            term_arguments = extract_terminal_arguments(extracted_appimage.desktop_entry.getExec())
            exec_arguments = term_arguments['arguments']
            el.exec_arguments = ' '.join(exec_arguments)

            jd_desktop_entry = JdDesktopEntry.from_file(
                extracted_appimage.desktop_file.get_path()
            )

            jd_desktop_entry.TryExec = dest_appimage_file.get_path()
            jd_desktop_entry.Exec = shlex.join([dest_appimage_file.get_path(), *exec_arguments])

            if el.update_logic is AppImageUpdateLogic.KEEP:
                v_version = self._get_app_version(extracted_appimage, return_hash=True)
                final_app_name = extracted_appimage.desktop_entry.getName() + f' ({v_version})'
                jd_desktop_entry.Name.default_text = final_app_name

            desktop_icon = f'applications-other'
            if dest_appimage_icon_file:
                desktop_icon = f"{dest_appimage_icon_file.get_path()}"

            jd_desktop_entry.Icon = desktop_icon

            for a, action in jd_desktop_entry.Actions.items():
                a_exec_args = extract_terminal_arguments(action.Exec)
                action.Icon = desktop_icon
                action.Exec = shlex.join([dest_appimage_file.get_path(), *a_exec_args['arguments']])

            if version:
                jd_desktop_entry.CustomKeys['X-AppImage-Version'] = version
            jd_desktop_entry.CustomKeys['X-AppImage-Name'] = extracted_appimage.desktop_entry.getName()
            desktop_file_content = jd_desktop_entry.get_text()

            # finally, write the new .desktop file
            if (not os.path.exists(self.user_desktop_files_path)) and os.path.exists(self.user_local_share_path):
                os.mkdir(self.user_desktop_files_path)

            with open(dest_desktop_file_path, 'w+') as desktop_file_python_dest:
                desktop_file_python_dest.write(desktop_file_content)

            if os.path.exists(dest_desktop_file_path):
                el.desktop_entry = DesktopEntry.DesktopEntry(filename=dest_desktop_file_path)
                el.desktop_file_path = dest_desktop_file_path
                el.installed_status = InstalledStatus.INSTALLED

            if el.updating_from and el.updating_from.env_variables:
                el.env_variables = el.updating_from.env_variables
                self.update_desktop_file(el)

            has_desktop_integration = False
            for v in el.env_variables:
                if v.startswith('DESKTOPINTEGRATION='):
                    has_desktop_integration = True
                    break

            if not has_desktop_integration:
                el.env_variables.append('DESKTOPINTEGRATION=1')
                self.update_desktop_file(el)

            Config.set_app_config(el, {})

        except Exception as e:
            logging.error('Appimage installation error: ' + str(e))
            raise e

        if Settings.settings.get_boolean('move-appimage-on-integration'):
            if os.path.dirname(extracted_appimage.appimage_file.get_path()) != appimages_destination_path:
                logging.info('Deleting original appimage file from: '  + extracted_appimage.appimage_file.get_path())
                if not extracted_appimage.appimage_file.delete(None):
                    raise InternalError('Cannot delete original file')

        update_dkt_db = terminal.host_sh(['update-desktop-database', self.user_desktop_files_path, '-q'], return_stderr=True)
        logging.debug(update_dkt_db)

        el.updating_from = None

    def reload_metadata(self, el: AppImageListElement):
        if not (el.installed_status is InstalledStatus.INSTALLED):
            return

        logging.info(f'Reloading metadata for {el.file_path}')
        random_str = get_random_string()
        dest_path = f'{self.extraction_folder}/gearlever_{random_str}'

        if not os.path.exists(dest_path):
            os.makedirs(dest_path)

        outdated_file = Gio.File.new_for_path(el.file_path)
        new_file = Gio.File.new_for_path(f'{dest_path}/tmp.appimage')

        gio_copy(outdated_file, new_file)

        self.uninstall(el, remove_configuration=False)

        el.file_path = f'{dest_path}/tmp.appimage'
        el.extracted = None

        self.install_file(el)

    def get_appimage_type(self, el: AppImageListElement) -> str:
        # https://github.com/AppImage/AppImageSpec/blob/fb05d9e1b8b8616dbeb7491303edc537dca573f3/draft.md#type-1-image-format

        appimage_type = 0
        with open(el.file_path, 'rb') as f:
            magic = f.read(11)[-3:]

            if magic == b'\x41\x49\x01':  # 0x414901
                appimage_type = 1
            elif magic == b'\x41\x49\x02':  # 0x414902
                appimage_type = 2

        return str(appimage_type)

    def create_list_element_from_file(self, file: Gio.File, return_new_el=False) -> AppImageListElement:
        if not self.can_install_file(file):
            raise InternalError(message='This file type is not supported')

        app_name: str = os.path.basename(file.get_parse_name())
        preview_enabled = False

        file_size = os.stat(file.get_path()).st_size
        el = AppImageListElement(
            name=re.sub(r'\.appimage$', '', app_name, 1, re.IGNORECASE),
            description='',
            version=None,
            provider=self.name,
            installed_status=InstalledStatus.NOT_INSTALLED,
            file_path=file.get_path(),
            size=file_size,
            desktop_entry=None,
            local_file=True,
            trusted=(not preview_enabled)
        )

        el.architecture = self.get_elf_arch(el)

        if return_new_el:
            return el

        if self.is_installed(el):
            for installed in self.list_installed():
                if filecmp.cmp(installed.file_path, el.file_path, shallow=False):
                    return installed

        return el

    def extraction_folder_cleanup(self):
        logging.debug(f'Clearing {self.extraction_folder}')
        if os.path.exists(self.extraction_folder):
            shutil.rmtree(self.extraction_folder)
            os.makedirs(self.extraction_folder)

    def update_exec_arguments(self, el:AppImageListElement, arg_string: str):
        arg_string = arg_string.replace("\n", "")

        if not el.desktop_file_path:
            raise Exception('desktop_file_path not specified')

        desktop_file_content = ''
        entry = DesktopEntry.DesktopEntry(filename=el.desktop_file_path)
        with open(el.desktop_file_path, 'r') as desktop_file:
            desktop_file_content = desktop_file.read()
            exec_command = shlex.split(entry.getExec())[0]
            exec_command += f' {arg_string}'

            # replace executable path
            desktop_file_content = re.sub(
                r'^Exec=.*$',
                f"Exec={exec_command}",
                desktop_file_content,
                flags=re.MULTILINE
            )

        with open(el.desktop_file_path, 'w') as desktop_file:
            desktop_file.write(desktop_file_content)

    def update_desktop_file(self, el: AppImageListElement):
        if not el.desktop_file_path:
            raise Exception('desktop_file_path not specified')

        jd_desktop_file = JdDesktopEntry.from_file(el.desktop_file_path)

        tryexec_command = jd_desktop_file.TryExec
        exec_arguments = el.exec_arguments
        env_vars = ' '.join(el.env_variables)

        if exec_arguments:
            exec_arguments = f' {exec_arguments}'

        if env_vars:
            env_vars = f'env {env_vars} '

        exec_command = ''.join([
            env_vars,
            shlex.quote(tryexec_command),
            exec_arguments
        ])

        jd_desktop_file.Exec = exec_command
        jd_desktop_file.write_file(el.desktop_file_path)

        el.desktop_entry = DesktopEntry.DesktopEntry(filename=el.desktop_file_path)

    def update_from_url(self, manager, el: AppImageListElement, status_cb: callable) -> AppImageListElement | None:
        try:
            update_file_path, f_hash = manager.download(status_cb)
        except DownloadInterruptedException as de:
            return None
        except Exception as e:
            raise e

        update_gfile = Gio.file_new_for_path(update_file_path)

        if not self.can_install_file(update_gfile):
            raise Exception(_('The downloaded file is not a valid appimage, please check if the provided URL is correct'))

        list_element = self.create_list_element_from_file(update_gfile, return_new_el=True)

        list_element.update_logic = AppImageUpdateLogic.REPLACE
        list_element.updating_from = el
        self.install_file(list_element)

        list_element.updating_from = None
        list_element.update_logic = None

        return list_element

    def is_app_running(self, el: AppImageListElement) -> bool: 
        if not el.file_path:
            return False

        running = terminal.host_sh(["ps", "-eo", "exe"])
        spl = running.split('\n')

        for l in spl:
            l = l.strip()

            if l == el.file_path:
                return True
            
        return False

    # Private methods
    def _run_filepath(self, el: AppImageListElement):
        is_nixos = re.search(r"^NAME=NixOS$", get_osinfo(), re.MULTILINE) != None

        if is_nixos:
            self._nixos_checks()
            terminal.host_threaded_sh(['appimage-run', el.file_path], callback=self._check_launch_output, return_stderr=True)
            return

        exec_args = []
        if el.desktop_entry:
            exec_args = shlex.split(el.desktop_entry.getExec())[1:]
            exec_args = [i for i in exec_args if i not in self.desktop_exec_codes]

        terminal.host_threaded_sh([el.file_path, *exec_args], callback=self._check_launch_output, return_stderr=True)

    def _run_from_desktopentry(self, el: AppImageListElement):
        is_nixos = re.search(r"^NAME=NixOS$", get_osinfo(), re.MULTILINE) != None

        if is_nixos:
            self._nixos_checks()
            cmd = ['appimage-run', el.desktop_entry.getTryExec()]
            return

        cmd = shlex.split(el.desktop_entry.getExec())
        cmd = [i for i in cmd if i not in self.desktop_exec_codes]

        terminal.host_threaded_sh(cmd, callback=self._check_launch_output, return_stderr=True)

    def _nixos_checks(self):
        try:
            terminal.host_sh(['which', 'appimage-run'])
        except Exception as e:
            msg = _("Running AppImages on NixOS requires appimage-run")
            raise Exception(msg)

    @idle
    def _check_launch_output(self, output: str):
        output = output.strip()

        if output:
            # Printing the output might help folks who run
            # run Gear Lever from the terminal
            print(output)

            if self.v2_detector_string in output:
                show_message_dialog(
                    _('AppImages require FUSE to run. You might still be able to run it with --appimage-extract-and-run in the command line arguments. \n\nClick the link below for more information. \n{url}'.format(
                        url='<a href="https://github.com/AppImage/AppImageKit/wiki/FUSE">https://github.com/AppImage/AppImageKit/wiki/FUSE</a>'
                    )),
                    markup=True
                )

    def _extract_appimage(self, el: AppImageListElement) -> str:
        random_str = get_random_string()
        dest_path = os.path.join(self.extraction_folder, f'gearlever_{random_str}')

        file = Gio.File.new_for_path(el.file_path)

        if not os.path.exists(f'{dest_path}'):
            os.makedirs(f'{dest_path}')

        logging.debug(f'Created temporary folder at {dest_path}')

        ###############################################################################

        # We use 7zip to extract the content of the appimage bundle
        # Old versions of Gear Lever used the --appimage-extract command, which is not
        # supported by all the appimage packages

        ###############################################################################

        squashfs_root_folder = os.path.join(dest_path, 'squashfs-root')
        use_appimage_extract = False
        use_7z = False
        use_unsquashfs = False
        use_dwarf = False
        block_unsafe_extractor = Settings.settings.get_boolean('block-unsafe-extractor')


        try:
            terminal.sandbox_sh(['dwarfsck', f'--input={file.get_path()}', '-q', '-detail=0', '--no-check'], error_quiet=True)
            use_dwarf = True
        except:
            use_7z = True
            logging.info('Filesystem is not dwarfsck')

        if use_dwarf:
            os.mkdir(squashfs_root_folder)
            logging.info(f'Exctracting with dwarfsextract to {squashfs_root_folder}')

            try:
                terminal.sandbox_sh(['dwarfsextract', f'--input={file.get_path()}', f'--output={squashfs_root_folder}',
                                        '--pattern=**.png','--pattern=**.svg', '--pattern=**.desktop', '--pattern=.DirIcon'])
            except Exception as e:
                logging.error('Extraction with dwarfsextract failed')
                logging.error(str(e))
                use_appimage_extract = True

        if use_7z:
            try:
                terminal.sandbox_sh(['7z', 't', file.get_path(), '-y', '-bso0', '-bsp0'])
                z7zoutput = terminal.sandbox_sh(['7z', 'x', file.get_path(), f'-o{squashfs_root_folder}', '-y', '-bso0', '-bsp0',
                                                    '-ir!*.png', '-ir!*.svg', '-ir!*.desktop',
                                                    '-xr!*.desktop*/', '-xr!*.svg*/', '-xr!*.png*/',
                                                    '-ir0!.DirIcon'
                                                ], cwd=dest_path)
                logging.debug('=== 7z log ===')
                logging.debug(z7zoutput)
                logging.debug(f'=== end 7z log ===')
            except Exception as e:

                if glob.glob('*.desktop', root_dir=squashfs_root_folder) and \
                    glob.glob('*.png', root_dir=squashfs_root_folder) or glob.glob('*.svg', root_dir=squashfs_root_folder):
                    logging.warning('Extraction with 7z failed, but the required files have been extracted')
                else:
                    logging.error('Extraction with 7z failed')
                    logging.error(str(e))
                    use_unsquashfs = True

        if  use_unsquashfs:
            logging.debug('Testing with unsquashfs')
            appimage_offset = terminal.sandbox_sh(['/usr/share/gearlever/get_appimage_offset', file.get_path()])

            try:
                terminal.sandbox_sh(['unsquashfs', '-o', appimage_offset, '-l', file.get_path()])
                terminal.sandbox_sh(['unsquashfs', '-o', appimage_offset, '-d', squashfs_root_folder, file.get_path()])
            except Exception as e:
                logging.error('Extraction with unsquashfs failed')
                logging.error(str(e))
                use_appimage_extract = True

        if block_unsafe_extractor and use_appimage_extract:
            raise InternalError(
                _("Any attempts at loading this app's metadata failed; if you want proceed using appimage-extract, please disable the security option in the settings.")
            )

            raise Exception('blocked by security setting')

        if (block_unsafe_extractor == False) and use_appimage_extract:
            logging.info('Extracting with appimage-extract')
            cloned_file = Gio.File.new_for_path(f'{dest_path}/app.appimage')
            gio_copy(file, cloned_file)
            terminal.sandbox_sh(['chmod', '+x', cloned_file.get_path()])
            terminal.sandbox_sh([cloned_file.get_path(), '--appimage-extract'], cwd=dest_path)

        return squashfs_root_folder

    def _load_appimage_metadata(self, el: AppImageListElement) -> ExtractedAppImage:
        if el.extracted:
            return el.extracted

        file = Gio.File.new_for_path(el.file_path)

        icon_file: Optional[Gio.File] = None
        desktop_file: Optional[Gio.File] = None
        desktop_entry: Optional[DesktopEntry.DesktopEntry] = None

        # hash file
        md5_hash = get_file_hash(file)
        temp_file_name = 'gearlever_appimage_' + md5_hash
        tmp_folder = Gio.File.new_for_path(f'{self.extraction_folder}/{temp_file_name}')

        if tmp_folder.query_exists():
            shutil.rmtree(tmp_folder.get_path())

        if tmp_folder.make_directory_with_parents(None):
            mounted_appimage_path = self._extract_appimage(el)
            extraction_folder = Gio.File.new_for_path(mounted_appimage_path)

            try:
                if not extraction_folder.query_exists():
                    raise InternalError('Missing mounted extraction folder ' + mounted_appimage_path)

                for d in  os.listdir(f'{extraction_folder.get_path()}'):
                    if not d.endswith('.desktop'):
                        continue

                    gdesk_file = Gio.File.new_for_path(f'{extraction_folder.get_path()}/{d}')
                    if get_giofile_content_type(gdesk_file) == 'application/x-desktop':
                        desktop_file = Gio.File.new_for_path(f'{tmp_folder.get_path()}/app.desktop')
                        gio_copy(file=gdesk_file, destination=desktop_file)

                        break

                desktop_entry_icon = None
                if desktop_file:
                    desktop_entry = DesktopEntry.DesktopEntry(desktop_file.get_path())
                    desktop_entry_icon = desktop_entry.getIcon()
                    desktop_entry_icon = re.sub(r"\.(png|svg)$", '', desktop_entry_icon)

                if desktop_entry_icon:
                    # https://github.com/AppImage/AppImageSpec/blob/master/draft.md#the-filesystem-image

                    tmp_icon_file: Optional[Gio.File] = None
                    
                    if not tmp_icon_file:
                        # if icon file is still not found, let's try with .DirIcon file
                        diricon = Gio.File.new_for_path(
                            os.path.join(extraction_folder.get_path(), '.DirIcon')
                        )

                        if diricon.query_exists():
                            diricon_ct = get_giofile_content_type(diricon)

                            # https://docs.appimage.org/reference/appdir.html
                            if diricon_ct in ['image/png', 'image/svg', 'image/svg+xml']:
                                tmp_icon_file = diricon
                            elif diricon_ct in ['text/plain']:
                                with open(diricon.get_path(), 'r') as f:
                                    possible_icon_path = f.read()
                                    diricon_linked_to = Gio.File.new_for_path(possible_icon_path)

                                    if diricon_linked_to.query_exists() and \
                                        get_giofile_content_type(diricon_linked_to) in ['image/png']:
                                        tmp_icon_file = diricon_linked_to
                    
                    if not tmp_icon_file:
                        icon_xt_f = None
                        for icon_xt in ['.svg', '.png']:
                            icon_xt_f = Gio.File.new_for_path(extraction_folder.get_path() + f'/{desktop_entry_icon}{icon_xt}')

                            if icon_xt_f.query_exists():
                                tmp_icon_file = icon_xt_f
                                break

                        if (not icon_xt_f.query_exists()) or (get_giofile_content_type(icon_xt_f) not in ['image/svg+xml', 'image/svg']):
                            # always prefer svg(s) to png(s)
                            # if a png is not found in the root of the filesystem, try somewhere else

                            icons_folder_prefix = '/usr/share/icons/hicolor'
                            icon_try_paths = [
                                extraction_folder.get_path() + f'{icons_folder_prefix}/scalable/apps/{desktop_entry_icon}.svg',
                                extraction_folder.get_path() + f'{icons_folder_prefix}/512x512/apps/{desktop_entry_icon}.png',
                                extraction_folder.get_path() + f'{icons_folder_prefix}/256x256/apps/{desktop_entry_icon}.png',
                                extraction_folder.get_path() + f'{icons_folder_prefix}/128x128/apps/{desktop_entry_icon}.png',
                                extraction_folder.get_path() + f'{icons_folder_prefix}/96x96/apps/{desktop_entry_icon}.png'
                            ]

                            for icon_xt in icon_try_paths:
                                logging.debug('Looking for icon in: ' + icon_xt)
                                icon_xt_f = Gio.File.new_for_path(icon_xt)

                                if icon_xt_f.query_exists():
                                    tmp_icon_file = icon_xt_f
                                    break

                    if tmp_icon_file:
                        i, tmp_icon_ext = os.path.splitext(tmp_icon_file.get_path())
                        icon_file = Gio.File.new_for_path(f'{tmp_folder.get_path()}/icon{tmp_icon_ext}')
                        gio_copy(file=tmp_icon_file, destination=icon_file)

            except Exception as e:
                logging.error(str(e))

        result = ExtractedAppImage()
        result.desktop_entry = desktop_entry
        result.extraction_folder = tmp_folder.get_path()
        result.appimage_file = file
        result.desktop_file = desktop_file
        result.icon_file = icon_file
        result.md5 = md5_hash

        el.desktop_entry = desktop_entry
        el.extracted = result

        return result

    def _get_appimages_default_destination_path(self) -> str:
        folder = Settings.settings.get_string('appimages-default-folder')
        return re.sub(r'^~', GLib.get_home_dir(), folder)

    def _get_app_version(self, extracted_appimage: Optional[ExtractedAppImage], desktop_entry: Optional[DesktopEntry] = None, return_hash=True):
        if not desktop_entry:
            desktop_entry = extracted_appimage.desktop_entry

        version = desktop_entry.get('X-AppImage-Version', '')

        if (not version) and extracted_appimage and return_hash:
            version = extracted_appimage.md5[0:6]

        return version

    def get_elf_arch(self, el: AppImageListElement) -> AppImageArchitecture:
        file_brief = terminal.sandbox_sh(['file', '--brief', '--exclude-quiet=apptype', '--exclude-quiet=ascii',
                                                     '--exclude-quiet=compress', '--exclude-quiet=csv', '--exclude-quiet=elf',
                                                     '--exclude-quiet=encoding', '--exclude-quiet=tar', '--exclude-quiet=cdf',
                                                     '--exclude-quiet=json', '--exclude-quiet=simh', '--exclude-quiet=text', '--exclude-quiet=tokens',
                                                     el.file_path])

        aarch = AppImageArchitecture.UNKNOWN
        file_brief = file_brief.lower()

        if 'aarch64' in file_brief or ' arm ' in file_brief:
            aarch = AppImageArchitecture.ARM_64
        elif 'x86-64' in file_brief or 'x86_64' in file_brief:
            aarch = AppImageArchitecture.X86_64

        return aarch



import logging
import requests
import os
from fnmatch import fnmatch
from typing import Optional
from urllib.parse import urlsplit
from gi.repository import Adw

from ..lib import json_config
from ..lib.ini_config import Config
from .UpdateManager import UpdateManager
from .StaticFileUpdater import StaticFileUpdater
from ..components.AdwEntryRowDefault import AdwEntryRowDefault

# Example:
# https://codeberg.org/sonusmix/sonusmix/

class CodebergUpdater(UpdateManager):
    staticfile_manager: Optional[StaticFileUpdater]
    label = 'Codeberg'
    name = 'CodebergUpdater'

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.staticfile_manager = None
        self.embedded = False
        self.repo_url_row = None
        self.repo_filename_row = None
        self.allow_prereleases_row = None

    def get_url_data(self, url: str):
        if url.startswith('https://'):
            logging.debug(f'CodebergUpdater: found http url, trying to detect codeberg data')
            urldata = urlsplit(url)

            if urldata.netloc != 'codeberg.org':
                return None

            paths = urldata.path.split('/')

            if len(paths) < 3:
                return None

            return {
                'username': paths[1],
                'repo': paths[2],
            }

        return None

    def migrate_v2(self):
        app_config = json_config.read_config_for_app(self.el)
        config = None

        if 'update_url' in app_config:
            urldata = urlsplit(app_config['update_url'])
            paths = urldata.path.split('/')

            # Old download URLs have the form:
            # https://codeberg.org/username/repo/releases/download/tag/filename
            if urldata:
                config = {
                    'allow_prereleases': app_config.get('update_manager_config', {}).get('allow_prereleases', False),
                    'repo': f'{paths[1]}/{paths[2]}',
                    'repo_filename': paths[6],
                }

        if config:
            Config.set_app_update_config(self.el, self, config)

    def download(self, status_update_cb) -> tuple[str, str]:
        target_asset = self.fetch_target_asset()
        if not target_asset:
            raise Exception(f'Missing target_asset for {self.name} instance')

        dwnl = target_asset['browser_download_url']
        self.staticfile_manager = StaticFileUpdater(self.el, dwnl)
        fname, etag = self.staticfile_manager.download(status_update_cb)

        self.staticfile_manager = None
        return fname, target_asset['id']

    def cancel_download(self):
        if self.staticfile_manager:
            self.staticfile_manager.cancel_download()
            self.staticfile_manager = None

    def cleanup(self):
        if self.staticfile_manager:
            self.staticfile_manager.cleanup()

    def fetch_target_asset(self):
        conf = self.get_config()
        url_data = conf.get('repo', '').split('/')

        if len(url_data) < 2:
            return

        allow_prereleases = conf.getboolean('allow_prereleases', False)

        rel_url = '/'.join([
            'https://codeberg.org/api/v1/repos',
            url_data[0],
            url_data[1],
            'releases'
        ])

        if allow_prereleases:
            rel_url += '?draft=exclude'
        else:
            rel_url += '?pre-release=exclude&draft=exclude'

        try:
            rel_data_resp = requests.get(rel_url)
            rel_data_resp.raise_for_status()
            rel_data = rel_data_resp.json()
        except Exception as e:
            logging.error(e)
            return

        logging.debug(f'Found {len(rel_data)} releases from {rel_url}')
        if not rel_data:
            return

        if not conf.get('repo_filename'):
            return

        download_asset = None
        possible_targets = []
        for asset in rel_data[0]['assets']:
            if fnmatch(asset['name'], conf.get('repo_filename', '')):
                possible_targets.append(asset)

        if len(possible_targets) == 1:
            download_asset = possible_targets[0]
        else:
            logging.info(f'found {len(possible_targets)} possible file targets')

            for t in possible_targets:
                logging.info(' - ' + t['name'])

            if self.system_arch == 'x86_64':
                for t in possible_targets:
                    if self.is_x86.search(t['name']) or not self.is_arm.search(t['name']):
                        download_asset = t
                        logging.info('found possible target: ' + t['name'])
                        break

        if not download_asset:
            logging.debug(f'No matching assets found from {rel_url}')
            return

        logging.debug(f'Found 1 matching asset: {download_asset["browser_download_url"]}')
        return download_asset

    def is_update_available(self):
        config = self.get_config()
        target_asset = self.fetch_target_asset()

        if target_asset:
            old_size = os.path.getsize(self.el.file_path)
            is_size_different = target_asset['size'] != old_size
            return is_size_different

        if config.get('repo_filename'):
            return None

        return False

    def load_form_rows(self, embedded=None):
        config = self.get_config()
        repo_url = config.get('repo')
        filename = config.get('repo_filename')

        self.repo_url_row = AdwEntryRowDefault(
            text=repo_url,
            icon_name='gl-git',
            sensitive=True,
            title=('Username/Repo')
        )

        self.repo_filename_row = AdwEntryRowDefault(
            text=filename,
            icon_name='gl-paper',
            sensitive=True,
            title=_('Release file name')
        )

        self.allow_prereleases_row = Adw.SwitchRow(
            title=_('Allow pre-releases'),
            active=config.getboolean('allow_prereleases', False)
        )

        return [
            self.repo_url_row,
            self.repo_filename_row,
            self.allow_prereleases_row
        ]

    def get_config_from_form(self):
        allow_prereleases = False
        repo_url = None
        repo_filename = None

        if self.allow_prereleases_row:
            allow_prereleases = self.allow_prereleases_row.get_active()

        if self.repo_url_row:
            repo_url = self.repo_url_row.get_text().strip()

        if self.repo_filename_row:
            repo_filename = self.repo_filename_row.get_text().strip()

        return {
            'allow_prereleases': allow_prereleases,
            'repo': repo_url,
            'repo_filename': repo_filename,
        }

    def validate_config(self, config):
        if (len(config.get('repo', '').split('/'))) != 2:
            raise Exception(f'Invalid data, please enter <username>/<repo>')



import logging
import requests
import shutil
import os
import posixpath
from  urllib.parse import urlparse
import re
from gi.repository import Adw, Gio
from typing import Optional, Literal

from ..lib.utils import get_random_string, url_is_valid, get_file_hash
from ..lib import json_config
from ..lib.ini_config import Config
from ..providers.AppImageProvider import AppImageProvider, AppImageListElement
from .Models import DownloadInterruptedException
from ..components.AdwEntryRowDefault import AdwEntryRowDefault

from .UpdateManager import UpdateManager

class StaticFileUpdater(UpdateManager):
    label = _('Static URL')
    handles_embedded = 'zsync|'
    name = 'StaticFileUpdater'
    currend_download: Optional[requests.Response]

    @staticmethod
    def get_url_headers(url):
        headers = {}
        head_request_error = False

        try:
            resp = requests.head(url, allow_redirects=True)
            resp.raise_for_status()
            headers = resp.headers
        except Exception as e:
            head_request_error = True
            logging.error(str(e))
            
        if head_request_error:
            # If something goes wrong with the Head request, try with stream mode
            logging.warn('Head request failed, trying with stream mode...')

            try:
                resp = requests.get(url, allow_redirects=True, stream=True)
                with resp as r:
                    r.raise_for_status()
                    headers = r.headers
                    r.close()
            except Exception as e:
                logging.error(str(e))
        
        return headers


    def __init__(self, el, embedded=False) -> None:
        super().__init__(embedded=embedded, el=el)
        self.form_row = None

    def download(self, status_update_cb) -> tuple[str, str]:
        random_name = get_random_string()
        fname = f'{self.download_folder}/{random_name}.appimage'

        if not os.path.exists(self.download_folder):
            os.makedirs(self.download_folder)

        dwnl_url = self.get_config().get('url')
        edwnl_url = self.get_embedded_url()

        if edwnl_url:
            zsync_file = requests.get(edwnl_url).text
            zsync_file_header = zsync_file.split('\n\n', 1)[0]
            url_pattern = r"URL:\s(.*)"
            match = re.search(url_pattern, zsync_file_header)

            if match:
                zsyncfile_url = match.group(1)

                if zsyncfile_url.startswith('https://') or \
                    zsyncfile_url.startswith('http://'):
                    dwnl_url = zsyncfile_url
                else:
                    urlparsed = urlparse(edwnl_url)
                    pp = posixpath.join(posixpath.dirname(urlparsed.path), zsyncfile_url)
                    dwnl_url = urlparsed._replace(path=pp,query='',fragment='').geturl()
            else:
                dwnl_url = re.sub(r"\.zsync$", "", edwnl_url)

        if not dwnl_url:
            raise Exception('Missing download URL')

        self.currend_download = requests.get(dwnl_url, stream=True)
        etag = self.currend_download.headers.get("etag", '')
        total_size = int(self.currend_download.headers.get("content-length", 0))
        status = 0
        block_size = 1024

        if os.path.exists(fname):
            os.remove(fname)

        d_notify_at = 0.1
        with open(fname, 'wb') as f:
            for chunk in self.currend_download.iter_content(block_size):
                f.write(chunk)

                status += block_size

                if total_size:
                    d_perc = (status / total_size)

                    if d_perc > d_notify_at:
                        d_notify_at += 0.01
                        # logging.info(f'Download status {d_perc * 100}')
                        status_update_cb(status / total_size)

        if os.path.getsize(fname) < total_size:
            raise DownloadInterruptedException()

        self.currend_download = None
        return fname, etag

    def cancel_download(self):
        if self.currend_download:
            self.currend_download.close()
            self.currend_download = None

    def cleanup(self):
        if os.path.exists(self.download_folder):
            shutil.rmtree(self.download_folder)

    def get_embedded_url(self):
        if not self.embedded:
            return None
        
        if self.embedded.startswith('https://') or self.embedded.startswith('http://'):
            return self.embedded

        l = len(self.handles_embedded)
        return self.embedded[l:]

    def is_update_available(self):
        e_url = self.get_embedded_url()
        dwnl_url = self.get_config().get('url')

        if not self.el.file_path:
            return False

        if e_url:
            zsync_file = requests.get(e_url).text
            zsync_file_header = zsync_file.split('\n\n', 1)[0]
            sha_pattern = r"SHA-1:\s*([0-9a-f]{40})"
            curr_version_hash = get_file_hash(Gio.File.new_for_path(self.el.file_path), alg='sha1')

            match = re.search(sha_pattern, zsync_file_header)
            if match:
                updatable = match.group(1) != curr_version_hash
                logging.info('SHA-1 detected, app updatable: ' + str(updatable))
                return updatable

        headers = StaticFileUpdater.get_url_headers(dwnl_url)
        resp_cl = int(headers.get('content-length', '0'))
        old_size = os.path.getsize(self.el.file_path)

        logging.debug(f'StaticFileUpdater: new url has length {resp_cl}, old was {old_size}')

        if resp_cl == 0:
            return False

        is_size_different = resp_cl != old_size
        return is_size_different

    def load_form_rows(self):
        url = self.get_config().get('url', '')

        if self.get_embedded_url():
            url = self.get_embedded_url()

        self.form_row = AdwEntryRowDefault(
            text=url,
            icon_name='gl-earth',
            title=_('Update URL'),
            sensitive=(not self.embedded)
        )

        return [self.form_row]

    def get_config_from_form(self):
        url = ''

        if self.form_row:
            url = self.form_row.get_text()

        return {
            'url': url.strip()
        }

    def migrate_v2(self):
        app_config = json_config.read_config_for_app(self.el)

        if app_config.get('update_url'):
            logging.info('Performing config migration from v1 to v2 for ' + self.el.file_path)
            Config.set_app_update_config(self.el, self, {
                'url':  app_config.get('update_url')
            })

    def validate_config(self, config):
        url = config.get('url', '')
        if (not url.startswith('http://')) and (not url.startswith('https://')):
            raise Exception('Enter a valid HTTP url')


import gi
import threading
import logging
import os

from .lib.constants import FETCH_UPDATES_ARG
from .models.Models import InternalError
from .models.Settings import Settings
from .lib.utils import portal
from .lib.ini_config import Config
from dbus import Array as DBusArray

gi.require_version('Gtk', '4.0')

from gi.repository import Adw, Gtk, Gio, GLib  # noqa

class Preferences(Adw.PreferencesWindow):
    def __init__(self, **kwargs) :
        super().__init__(**kwargs)

        self.settings = Settings.settings

        # page 1
        page1 = Adw.PreferencesPage()

        # general group
        general_preference_group = Adw.PreferencesGroup(name=_('General'))


        # default_location
        self.default_location_row = Adw.ActionRow(
            title=_('AppImage default location'),
            subtitle=self.settings.get_string('appimages-default-folder')
        )

        pick_default_localtion_btn = Gtk.Button(icon_name='gearlever-file-manager-symbolic', valign=Gtk.Align.CENTER)
        pick_default_localtion_btn.connect('clicked', self.on_default_localtion_btn_clicked)
        self.default_location_row.add_suffix(pick_default_localtion_btn)

        files_outside_folder_switch = self.create_boolean_settings_entry(
            _('Show integrated AppImages outside the default folder'),
            'manage-files-outside-default-folder',
            _('List AppImages that have been integrated into the system menu but are located outside the default folder')
        )

        general_preference_group.add(self.default_location_row)
        general_preference_group.add(files_outside_folder_switch)

        # updates management group
        updates_management_group = Adw.PreferencesGroup(name=_('Updates management'), title=_('Updates management'))
        autofetch_updates = self.create_boolean_settings_entry(
            _('Check updates on system startup'),
            'fetch-updates-in-background',
            _('Receive a notification when a new update is detected; updates will not be installed automatically')
        )

        updates_management_group.add(autofetch_updates)
        autofetch_updates.connect('notify::active', self.on_background_fetchupdates_changed)

        # file management group
        move_appimages_group = Adw.PreferencesGroup(name=_('File management'), title=_('File management'))
        move_appimages_row = Adw.ActionRow(
            title=_('Move AppImages into the destination folder'),
            subtitle=(_('Reduce disk usage'))
        )

        copy_appimages_row = Adw.ActionRow(
            title=_('Clone AppImages into the destination folder'),
            subtitle=(_('Keep the original file and create a copy in the destination folder'))
        )


        self.move_to_destination_check = Gtk.CheckButton(
            valign=Gtk.Align.CENTER,
            active=self.settings.get_boolean('move-appimage-on-integration')
        )

        self.copy_to_destination_check = Gtk.CheckButton(
            valign=Gtk.Align.CENTER,
            group=self.move_to_destination_check,
            active=(not self.settings.get_boolean('move-appimage-on-integration'))
        )

        move_appimages_row.add_prefix(self.move_to_destination_check)
        copy_appimages_row.add_prefix(self.copy_to_destination_check)

        move_appimages_group.add(move_appimages_row)
        move_appimages_group.add(copy_appimages_row)
        # move_appimages_group.add(exec_as_name_switch)

        self.move_to_destination_check.connect('toggled', self.on_move_appimages_setting_changed)
        self.copy_to_destination_check.connect('toggled', self.on_move_appimages_setting_changed)

        # naming conventions group
        nconvention_group = Adw.PreferencesGroup(name=_('Naming conventions'), title=_('Naming conventions'))
        exec_as_name_switch = self.create_boolean_settings_entry(
            _('Use executable name for integrated terminal apps'),
            'exec-as-name-for-terminal-apps',
            _('If enabled, apps that run in the terminal are renamed as their executable.\nYou would need to add the aforementioned folder to your $PATH manually.\n\nFor example, "golang_x86_64.appimage" will be saved as "go"')
        )

        nconvention_group.add(exec_as_name_switch)

        preview_minimal_ui = self.create_boolean_settings_entry(
            _('Preview new apps in a minimal UI'),
            'preview-minimal-ui',
        )
        
        block_appimage_extract = self.create_boolean_settings_entry(
            _('Use safe mechanisms to load metadata'),
            'block-unsafe-extractor',
            _('If enabled, the app will try to load an Appimage without using the built-in appimage-extract command.\nDisable only if you trust your sources: malicious apps could execute code when loading the metadata.')
        )

        general_preference_group.add(preview_minimal_ui)
        general_preference_group.add(block_appimage_extract)

        # debugging group
        debug_group = Adw.PreferencesGroup(name=_('Debugging'), title=_('Debugging'))
        debug_row = self.create_boolean_settings_entry(
            _('Enable debug logs'),
            'debug-logs',
            _('Increases log verbosity, occupying more disk space and potentially impacting performance.\nRequires a restart.')
        )

        debug_group.add(debug_row)

        page1.add(general_preference_group)
        page1.add(updates_management_group)
        page1.add(move_appimages_group)
        page1.add(nconvention_group)
        page1.add(debug_group)
        self.add(page1)

    def on_select_default_location_response(self, dialog, result):
        file_path = ''

        try:
            selected_file = dialog.select_folder_finish(result)
            file_path = selected_file.get_path()
        except Exception as e:
            logging.error(str(e))
            return

        if selected_file.query_exists() and os.access(file_path, os.W_OK):
            self.settings.set_string('appimages-default-folder', file_path)
            self.default_location_row.set_subtitle(file_path)
        else:
            raise InternalError(_('The folder must writeable'))

    def on_default_localtion_btn_clicked(self, widget):
        dialog = Gtk.FileDialog(title=_('Select a folder'), modal=True)

        dialog.select_folder(
            parent=self,
            cancellable=None,
            callback=self.on_select_default_location_response
        )

    def on_move_appimages_setting_changed(self, widget):
        self.settings.set_boolean('move-appimage-on-integration', self.move_to_destination_check.get_active())

    def create_boolean_settings_entry(self, label: str, key: str, subtitle: str = None) -> Adw.SwitchRow:
        row = Adw.SwitchRow(title=label, subtitle=subtitle)
        self.settings.bind(key, row, 'active', Gio.SettingsBindFlags.DEFAULT)

        return row
        
    def on_background_fetchupdates_changed(self, *args):
        key = 'fetch-updates-in-background'
        value: bool = self.settings.get_boolean(key)

        Config.parser[Config.parser.default_section][key] = Config.return_boolean(value)
        Config.write()
        
        inter = portal("org.freedesktop.portal.Background")
        inter.RequestBackground('', {
            'reason': 'Gear Lever background updates fetch', 
            'autostart': value, 
            'background': True, 
            'commandline': DBusArray(['gearlever', f'--{FETCH_UPDATES_ARG}'])
        })



import random
import dbus
import re
import os
import time
import logging
import shlex
import gi
import hashlib
import requests

gi.require_version('Gtk', '4.0')
gi.require_version('Adw', '1')

from .async_utils import idle
from gi.repository import Gtk, Gio, Adw, Gdk, GLib, GdkPixbuf  # noqa


def key_in_dict(_dict: dict, key_lookup: str, separator='.'):
    """
        Searches for a nested key in a dictionary and returns its value, or None if nothing was found.
        key_lookup must be a string where each key is deparated by a given "separator" character, which by default is a dot
    """
    keys = key_lookup.split(separator)
    subdict = _dict

    for k in keys:
        if isinstance(subdict, dict):
            subdict = subdict[k] if (k in subdict) else None

        if subdict is None:
            break

    return subdict


def add_page_to_adw_stack(stack: Adw.ViewStack, page: Gtk.Widget, name: str, title: str, icon: str):
    stack.add_titled(page, name, title)
    stack.get_page(page).set_icon_name(icon)


# def gtk_image_from_url(url: str, image: Gtk.Image) -> Gtk.Image:
#     response = requests.get(url, timeout=10)
#     response.raise_for_status()

#     loader = GdkPixbuf.PixbufLoader()
#     loader.write_bytes(GLib.Bytes.new(response.content))
#     loader.close()

#     image.clear()
#     image.set_from_pixbuf(loader.get_pixbuf())
#     return image


def set_window_cursor(cursor: str):
    for w in Gtk.Window.list_toplevels():
        if isinstance(w, Gtk.ApplicationWindow):
            w.set_cursor(Gdk.Cursor.new_from_name(cursor, None))
            break


def get_application_window() -> Adw.Window:
    for w in Gtk.Window.list_toplevels():
        if isinstance(w, Adw.Window):
            return w


def get_giofile_content_type(file: Gio.File):
    return file.query_info('standard::', Gio.FileQueryInfoFlags.NONE, None).get_content_type()


def gio_copy(file: Gio.File, destination: Gio.File):
    logging.debug(f'Copying {file.get_path()} to {destination.get_path()}')
    return file.copy(
        destination,
        Gio.FileCopyFlags.OVERWRITE,
        None, None, None, None
    )


def get_file_hash(file: Gio.File | None, alg='md5', file_path=None) -> str:
    if not file_path:
        file_path = file.get_path()

    with open(file_path, 'rb') as f:
        if alg == 'md5':
            return hashlib.md5(f.read()).hexdigest()
        elif alg == 'sha1':
            return hashlib.sha1(f.read()).hexdigest()
        elif alg == 'sha256':
            return hashlib.sha256(f.read()).hexdigest()

    raise Exception('Invalid hash requested')


def send_notification(notification=Gio.Notification, tag=None):
    if not tag:
        tag = str(time.time_ns())
    Gio.Application().get_default().send_notification(tag, notification)


def create_dict(*args: str):
    return dict({i: eval(i) for i in args})


def portal(interface: str, bus_name: str='org.freedesktop.portal.Desktop', object_path: str='/org/freedesktop/portal/desktop') -> dbus.Interface:
    bus = dbus.SessionBus()
    obj = bus.get_object(bus_name, object_path)
    inter = dbus.Interface(obj, interface)

    return inter


def get_element_without_overscroll(arr: list, index: int):
    """
    Returns the element at the given index in the array.
    If the index is out of bounds, the index is wrapped around
    to the range of valid indices for the array.
    """
    if index < 0:
        index = 0

    if len(arr) == 0:
        raise ValueError("Array must not be empty")
    wrapped_index = index % len(arr)
    return arr[wrapped_index]


def url_is_valid(url: str) -> bool:
    url_regex = re.compile(
        r'^(?:http|https)://'  # http:// or https://
    , re.IGNORECASE)

    is_valid = True if url_regex.match(url) else False

    if not is_valid:
        logging.warn(f'Provided url "{url}" is not a valid url')

    return is_valid

def get_random_string():
    return ''.join((random.choice('abcdxyzpqr123456789') for i in range(10)))

def remove_special_chars(filename, replacement=""):
    """Removes special characters from a filename and replaces them with a chosen character.

    Args:
        filename: The filename to be sanitized.
        replacement: The character to replace special characters with (default: "_").

    Returns:
        The sanitized filename.
    """
    # Regular expression to match special characters (excluding alphanumeric, underscore, and dot)
    pattern = r"[^\w\._]+"
    return re.sub(pattern, replacement, filename)

@idle
def show_message_dialog(message, header=None, markup=False):
    dialog = Adw.MessageDialog(
        transient_for=get_application_window(),
        heading=header or _('Error'),
        body_use_markup=markup
    )

    dialog.set_body(message)

    dialog.set_body_use_markup(True)
    dialog.add_response('okay', _('Dismiss'))

    dialog.present()


def get_osinfo():
    os_release_file = "/run/host/os-release"
    if os.environ.get('FLATPAK_ID', None) is None:
        os_release_file = "/etc/os-release"

    output = ''
    if os.path.exists(os_release_file):
        with open(os_release_file, 'r') as f:
            output = f.read()
    return output

# thank you mate ❤️
# https://github.com/gtimelog/gtimelog/blob/6e4b07b58c730777dbdb00b3b85291139f8b10aa/src/gtimelog/main.py#L159
def make_option(long_name, short_name=None, flags=0, arg=0, arg_data=None, description=None, arg_description=None):
    # surely something like this should exist inside PyGObject itself?!
    option = GLib.OptionEntry()
    option.long_name = long_name.lstrip('-')
    option.short_name = 0 if not short_name else short_name.lstrip('-')
    option.flags = flags
    option.arg = int(arg)
    option.arg_data = arg_data
    option.description = description
    option.arg_description = arg_description
    return option


def avoid_file_conflicts(filename, directory):
    i = 0
    files_in_dest_dir = os.listdir(directory)
    
    name, ext = os.path.splitext(filename)

    # if there is already an app with the same name, 
    # we try not to overwrite
    while filename in files_in_dest_dir:
        filename = name + f'_{i}'

        if ext:
            filename = filename + ext

    return filename

def extract_terminal_arguments(command):
    """
    Extract all terminal arguments from a command string.
    Handles environment variables, quoted paths, and flags.
    """
    
    # Parse the command using shlex to handle quotes and escaping properly
    tokens = shlex.split(command)
    
    result = {
        'env_vars': [],
        'executable': '',
        'arguments': []
    }
    
    i = 0
    while i < len(tokens):
        token = tokens[i]
        
        # Check if it's an environment variable (starts with 'env' or has '=' in it)
        if token == 'env':
            i += 1
            continue
        elif '=' in token and not token.startswith('/') and not token.startswith('-'):
            # This is an environment variable
            result['env_vars'].append(token)
        elif not result['executable'] and not token.startswith('-'):
            # This is the executable (first non-flag, non-env-var token)
            result['executable'] = token
        else:
            # This is an argument
            result['arguments'].append(token)
        
        i += 1
    
    return result

def gnu_naturalsize(value, precision=1):
    """
    Format a number of bytes like humanize.naturalsize(value, gnu=True).
    Uses powers of 1024 and single-letter suffixes.
    """
    if value < 0:
        return f"-{gnu_naturalsize(abs(value), precision)}"
    
    # Standard GNU suffixes
    suffixes = ('B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB')
    
    # Set the base for binary (1024)
    base = 1024.0
    
    if value < base:
        return f"{value}B"
    
    # Calculate the magnitude
    import math
    i = int(math.floor(math.log(value, base)))
    
    # Handle cases where value might exceed our suffix list
    if i >= len(suffixes):
        i = len(suffixes) - 1
        
    v = value / math.pow(base, i)
    
    # Return formatted string (e.g., 953.7M)
    return f"{v:.{precision}f} {suffixes[i]}"

def check_internet(timeout=3):
    urls = [
        'https://fedoraproject.org/static/hotspot.txt',
        'http://fedoraproject.org/static/hotspot.txt',
    ]

    for url in urls:
        try:
            r = requests.get(url=url, timeout=timeout, allow_redirects=False)
            if r.status_code < 400:
                return True
        except Exception:
            continue

    logging.warning("No internet connection detected")
    return False



import logging
import re
from typing import Optional, Callable

from ..lib.constants import TMP_DIR
from ..lib import terminal
from ..lib.ini_config import Config
from ..providers.AppImageProvider import AppImageListElement

from .UpdateManager import UpdateManager
from .GithubUpdater import GithubUpdater
from .GitlabUpdater import GitlabUpdater
from .CodebergUpdater import CodebergUpdater
from .StaticFileUpdater import StaticFileUpdater
from .FTPUpdater import FTPUpdater
from .ForgejoUpdater import ForgejoUpdater

class UpdateManagerChecker():
    @staticmethod
    def get_models() -> list[UpdateManager]:
        return [StaticFileUpdater, GithubUpdater, GitlabUpdater, CodebergUpdater, FTPUpdater, ForgejoUpdater]

    @staticmethod
    def get_model_by_name(manager_label: str) -> UpdateManager:
        item = list(filter(lambda m: m.name == manager_label, 
                                    UpdateManagerChecker.get_models()))

        if not item:
            raise Exception('Invalid model name: ' + manager_label)

        return item[0]

    @staticmethod
    def check_url_for_app(el: AppImageListElement) -> Optional[UpdateManager]:
        app_conf = Config.get_app_update_config(el)
        update_url_manager: str | None = app_conf.get('manager', None)
        embedded_url = None

        model = None
        if update_url_manager:
            model = UpdateManagerChecker.get_model_by_name(update_url_manager)
        else:
            models = UpdateManagerChecker.get_models()
            embedded_app_data = UpdateManagerChecker.check_app_embedded_url(el)

            if embedded_app_data:
                for m in models:
                    if m.handles_embedded and \
                        embedded_app_data.startswith(m.handles_embedded):

                        logging.debug(f'Checking embedded url with {m.__name__}')

                        model = m
                        embedded_url = embedded_app_data
                        break

        if model:
            return model(embedded=embedded_url, el=el)
        return None

    @staticmethod
    def check_app_embedded_url(el: AppImageListElement) -> Optional[str]:
        readelf_out = terminal.sandbox_sh(['readelf', '--string-dump=.upd_info', '--wide', el.file_path])
        readelf_out = readelf_out.replace('\n', ' ') + ' '

        # Github url
        # example value: " String dump of section '.upd_info':   [     0]  gh-releases-zsync|neovim|neovim|latest|nvim-linux-x86_64.appimage.zsync "
        pattern_gh = r"gh-releases-zsync\|.*(.zsync)"
        matches = re.search(pattern_gh, readelf_out)

        if matches:
            return matches[0].strip()

        # Static url
        # example value: " String dump of section '.upd_info':   [     0]  zsync|https://gitlab.com/api/v4/projects/24386000/packages/generic/librewolf/latest/LibreWolf.x86_64.AppImage.zsync "
        pattern_link = r"\szsync\|http(.*)\s"
        matches = re.search(pattern_link, readelf_out)

        if matches:
            return matches[0].strip()

        return None




from gi.repository import Gtk, Pango, GObject, Gio, GdkPixbuf, GLib, Adw


class FilterEntry(Gtk.SearchBar):
    def __init__(self, label, capture: Gtk.Widget = None, maximum_size=600):
        super().__init__()
    
        self.search_entry = Gtk.SearchEntry(placeholder_text=label, hexpand=True)
        self.search_entry.get_first_child().set_from_icon_name("gl-funnel-outline-symbolic")

        clamp = Adw.Clamp(child=self.search_entry, maximum_size=maximum_size,hexpand=True, margin_top=5, margin_bottom=5)

        self.set_search_mode(False)
        self.set_key_capture_widget(capture)
        self.set_child(clamp)

        self.connect_entry(self.search_entry)
        self.props.margin_bottom = 5



from urllib import request
from gi.repository import Gtk, Pango, GObject
from typing import Dict, List, Optional
from ..lib.async_utils import idle, _async
from ..lib.utils import gnu_naturalsize

from ..models.AppListElement import InstalledStatus
from ..providers.AppImageProvider import AppImageListElement
from ..providers.providers_list import appimage_provider


class AppListBoxItem(Gtk.ListBoxRow):
    __gsignals__ = {
        "details-clicked": (GObject.SIGNAL_RUN_FIRST, GObject.TYPE_NONE, (object, )),
    }

    def __init__(self, list_element: AppImageListElement, show_details_btn=False, **kwargs):
        super().__init__(**kwargs)

        self._app: AppImageListElement = list_element
        self.details_btn: Optional[Gtk.Button] = None

        col = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=5)
        col.set_css_classes(['app-listbox-item'])

        self.image_container = Gtk.Box()
        col.append(self.image_container)

        app_details_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, valign=Gtk.Align.CENTER)
        app_details_box.append(
            Gtk.Label(
                label=list_element.name,
                halign=Gtk.Align.START,
                # use_markup=True,
                css_classes=['heading'],
                max_width_chars=70,
                ellipsize=Pango.EllipsizeMode.END
            )
        )

        if list_element.description:
            app_details_box.append(
                Gtk.Label(
                    label=list_element.description, 
                    halign=Gtk.Align.START,
                    lines=1,
                    max_width_chars=100, 
                    ellipsize=Pango.EllipsizeMode.END,
                )
            )

        self.update_version = Gtk.Label(
            label='',
            margin_top=3,
            halign=Gtk.Align.START,
            css_classes=['subtitle'],
        )

        app_details_box.append(self.update_version)
        app_details_box.set_hexpand(True)
        col.append(app_details_box)

        if show_details_btn:
            self.details_btn = Gtk.Button(icon_name='gl-right-symbolic',
                                     valign=Gtk.Align.CENTER)

            col.append(self.details_btn)

        self.update_available_btn = Gtk.Button(
            icon_name='gl-software-update-available-symbolic',
            valign=Gtk.Align.CENTER,
            css_classes=['flat'],
            sensitive=False,
            visible=False
        )

        col.append(self.update_available_btn)
        
        self.set_child(col)

        if self._app.installed_status in [InstalledStatus.UPDATING, InstalledStatus.INSTALLING]:
            self.set_opacity(0.5)

    def load_icon(self):
        image = appimage_provider.get_icon(self._app)
        self.set_icon(image)
    
    def set_icon(self, image: Gtk.Image):
        image.set_pixel_size(45)
        self.image_container.append(image)

    def set_update_version(self, text: Optional[str], size: int):
        c = []

        if text:
            c.append(text)
        c.append(gnu_naturalsize(size))
        

        self.update_version.set_visible(True)
        self.update_version.set_label(' · '.join(c))

    def show_updatable_badge(self):
        self.update_available_btn.set_visible(True)


import logging
import os
from gi.repository import Gtk, Adw, GObject, Gio, GLib
from typing import Dict, List, Optional


from .models.Settings import Settings
from .lib.utils import get_element_without_overscroll, gio_copy
from .lib.constants import APP_ID, APP_NAME, APP_DATA

class WelcomeScreen(Adw.Dialog):

    def __init__(self):
        super().__init__()

        pkgdatadir = APP_DATA['PKGDATADIR']

        self.set_content_height(600)
        self.set_content_width(600)

        toolbar_view = Adw.ToolbarView()

        self.titlebar = Adw.HeaderBar()
        toolbar_view.add_top_bar(self.titlebar)

        overlay = Gtk.Overlay()

        self.carousel = Adw.Carousel(spacing=200)
        self.carousel.connect('page-changed', self.on_page_changed)
        
        dots = Adw.CarouselIndicatorDots()
        dots.set_carousel(self.carousel)
        self.titlebar.set_title_widget(dots)

        overlay.set_child(self.carousel)

        self.left_button = Gtk.Button(icon_name='go-previous')
        self.left_button.add_css_class('circular')
        self.left_button.set_sensitive(False)
        self.left_button.set_halign(Gtk.Align.START)
        self.left_button.set_valign(Gtk.Align.CENTER)
        self.left_button.set_margin_start(20)

        self.right_button = Gtk.Button(icon_name='go-next')
        self.right_button.add_css_class('circular')
        self.right_button.set_halign(Gtk.Align.END)
        self.right_button.set_valign(Gtk.Align.CENTER)
        self.right_button.set_margin_end(20)

        overlay.add_overlay(self.left_button)
        overlay.add_overlay(self.right_button)

        first_page = Gtk.Builder.new_from_resource(f'/it/mijorus/{APP_NAME}/gtk/tutorial/1.ui')
        second_page = Gtk.Builder.new_from_resource(f'/it/mijorus/{APP_NAME}/gtk/tutorial/2.ui')
        third_page = Gtk.Builder.new_from_resource(f'/it/mijorus/{APP_NAME}/gtk/tutorial/3.ui')
        last_page = Gtk.Builder.new_from_resource(f'/it/mijorus/{APP_NAME}/gtk/tutorial/last.ui')

        pages = [el.get_object('target') for el in [first_page, second_page, third_page, last_page]]
        [self.carousel.append(el) for el in pages]

        location_label = second_page.get_object('location-label')
        location_label.set_label(location_label.get_label().replace('{location}', '~/AppImages'))
        last_page.get_object('close-window').connect('clicked', lambda w: self.close())

        self.left_button.connect('clicked', lambda w: self.carousel.scroll_to(get_element_without_overscroll(pages, int(self.carousel.get_position()) - 1), True))
        self.right_button.connect('clicked', lambda w: self.carousel.scroll_to(get_element_without_overscroll(pages, int(self.carousel.get_position()) + 1), True))

        toolbar_view.set_content(overlay)

        self.demo_folder = os.path.join(GLib.get_user_cache_dir(), APP_ID, 'demo')
        demo_app = Gio.File.new_for_path(os.path.join(pkgdatadir, APP_NAME, 'assets', 'demo.AppImage'))
        demo_app_dest = os.path.join(self.demo_folder, demo_app.get_basename())

        # if the demo file exists, the path to it exists so we can skip checking both
        if not os.path.exists(demo_app_dest):
            # create paths to demo file if the path does not exist
            os.makedirs(self.demo_folder, exist_ok=True)
            # move demo file to the demo_folder
            gio_copy(demo_app, Gio.File.new_for_path(demo_app_dest))
            logging.debug(f'Copied demo app into {self.demo_folder}')

        third_page.get_object('open-demo-folder').connect('clicked', self.on_open_demo_folder_clicked)
        second_page.get_object('open-preferences').connect('clicked', self.on_default_localtion_btn_clicked)

        self.set_child(toolbar_view)


    def on_page_changed(self, widget, index):
        self.left_button.set_sensitive(True)
        self.right_button.set_sensitive(True)

        if index == 0:
            self.left_button.set_sensitive(False)

        if index == (self.carousel.get_n_pages() - 1):
            self.right_button.set_sensitive(False)

    def on_open_demo_folder_clicked(self, widget):
        gfile = Gio.File.new_for_path(self.demo_folder)
        launcher = Gtk.FileLauncher.new(gfile)
        launcher.launch()

    def on_select_default_location_response(self, dialog, result):
        try:
            selected_file = dialog.select_folder_finish(result)
        except Exception as e:
            logging.error(str(e))
            return

        if selected_file.query_exists() and selected_file.get_path().startswith(GLib.get_home_dir()):
            Settings.settings.set_string('appimages-default-folder', selected_file.get_path())
            # state.set__('appimages-default-folder', selected_file.get_path())
        else:
            raise InternalError(_('The folder must be in your home directory'))

    def on_default_localtion_btn_clicked(self, widget):
        dialog = Gtk.FileDialog(title=_('Select a folder'), modal=True)

        dialog.select_folder(
            parent=self.get_root(),
            cancellable=None,
            callback=self.on_select_default_location_response
        )



# window.py
#
# Copyright 2022 Lorenzo Paderi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from .lib.constants import APP_ID
from .InstalledAppsList import InstalledAppsList
from .AppDetails import AppDetails
from .MultiInstall import MultiInstall
from .MultiUpdate import MultiUpdate
from .providers.providers_list import appimage_provider
from .providers.AppImageProvider import AppImageListElement
from .models.AppListElement import AppListElement
from .models.Settings import Settings
from .lib import utils

from gi.repository import Gtk, Adw, Gio, Gdk, GLib


class GearleverWindow(Adw.Window):
    def __init__(self, from_file=False, **kwargs):
        super().__init__(**kwargs)
        self.from_file = from_file
        self.selected_files_count = 0
        self.open_appimage_tooltip = _('Open a new AppImage')
        self.settings = Settings.settings

        # Create a container stack 
        self.container_stack = Adw.Leaflet(can_unfold=False, can_navigate_back=True, can_navigate_forward=False)

        # Create the "main_stack" widget we will be using in the Window
        self.app_lists_stack = Adw.ViewStack()

        self.titlebar = Adw.HeaderBar()
        self.view_title_widget = Adw.ViewSwitcherTitle(stack=self.app_lists_stack)
        self.open_appimage_button_child = Adw.ButtonContent(icon_name='gl-plus-symbolic', 
                                                            tooltip_text=self.open_appimage_tooltip, label=_('Open'))

        self.search_btn = Gtk.Button(child=Adw.ButtonContent(icon_name='gl-loupe-large'))
        self.search_btn.connect('clicked', self.on_trigger_search_mode)

        self.left_button = Gtk.Button(
            child=self.open_appimage_button_child
        )

        menu_obj = Gtk.Builder.new_from_resource('/it/mijorus/gearlever/gtk/main-menu.ui')
        self.menu_button = Gtk.MenuButton(icon_name='open-menu', menu_model=menu_obj.get_object('primary_menu'))

        self.titlebar.pack_start(self.left_button)
        self.titlebar.pack_end(self.menu_button)
        self.titlebar.pack_end(self.search_btn)
        
        self.titlebar.set_title_widget(self.view_title_widget)

        toolbar_view = Adw.ToolbarView()
        toolbar_view.add_top_bar(self.titlebar)

        self.set_title('Gear lever')
        self.set_default_size(700, 700)

        toast_overlay = Adw.ToastOverlay()

        # Create the "stack" widget for the "installed apps" view
        self.installed_stack = Gtk.Stack()
        self.app_details = AppDetails()
        self.app_details.connect('uninstalled-app', self.on_uninstalled_app)
        self.app_details.connect('update-started', self.on_app_update_started)
        self.app_details.connect('update-ended', self.on_app_update_ended)

        self.installed_apps_list = InstalledAppsList()
        self.installed_apps_list.refresh_list()
        self.installed_apps_list.connect('update-all', self.on_update_all_event)

        self.multi_install = MultiInstall()
        self.multi_install.connect('show-details', self.on_multi_install_show_details)
        self.multi_install.connect('go-back', self.on_left_button_clicked)

        self.multi_update = MultiUpdate()
        self.multi_update.connect('go-back', self.on_left_button_clicked)

        self.installed_stack.add_child(self.installed_apps_list)
        self.installed_stack.set_visible_child(self.installed_apps_list)
        
        # Add content to the main_stack
        utils.add_page_to_adw_stack(self.app_lists_stack, self.installed_stack, 
                                    'installed', 'Installed', 'gearlever-computer-symbolic' )

        self.container_stack.append(self.app_lists_stack)
        self.container_stack.append(self.app_details)
        self.container_stack.append(self.multi_install)
        self.container_stack.append(self.multi_update)

        # Show details of an installed app
        self.installed_apps_list.connect('selected-app', self.on_selected_installed_app)

        # left arrow click
        self.left_button.connect('clicked', self.on_left_button_clicked)
        # change visible child of the app list stack
        self.app_lists_stack.connect('notify::visible-child', self.on_app_lists_stack_change)
        # change visible child of the container stack
        self.container_stack.connect('notify::visible-child', self.on_container_stack_change)

        builder = Gtk.Builder.new_from_resource('/it/mijorus/gearlever/gtk/drag-drop.ui')
        self.drag_drop_ui = builder.get_object('drag-drop')

        self.drop_target_controller = Gtk.DropTarget.new(Gio.File, Gdk.DragAction.COPY)
        self.drop_target_controller.set_gtypes([Gdk.FileList])

        self.drop_target_controller.connect('drop', self.on_drop_event)
        self.drop_target_controller.connect('enter', self.on_drop_enter)
        self.drop_target_controller.connect('leave', self.on_drop_leave)
        self.visible_before_dragdrop_start = None

        self.container_stack.add_controller(self.drop_target_controller)
        self.container_stack.append(self.drag_drop_ui)

        self.connect('close-request', self.on_close_request)
        self.connect('notify::maximized', self.on_window_maximixed_changed)

        toast_overlay.set_child(self.container_stack)

        toolbar_view.set_content(toast_overlay)
        self.set_content(toolbar_view)

        if self.settings.get_boolean('is-maximized'):
            self.maximize()

    # Show app details
    def on_selected_installed_app(self, source: Gtk.Widget, list_element: AppListElement):
        self.selected_files_count = 0
        self.app_details.set_app_list_element(list_element)
        self.container_stack.set_transition_type(Adw.LeafletTransitionType.OVER)
        self.container_stack.set_visible_child(self.app_details)

    def on_selected_local_file(self, files: list[Gio.File]) -> bool:
        if files:
            self.container_stack.set_transition_type(
                Adw.LeafletTransitionType.UNDER if self.from_file else Adw.LeafletTransitionType.OVER
            )

        if len(files) == 1 and self.app_details.set_from_local_file(files[0]):
            self.container_stack.set_visible_child(self.app_details)

            if self.from_file:
                if self.settings.get_boolean('preview-minimal-ui'):
                    self.app_details.set_minimal_ui(True)
                    self.set_default_size(-1, -1)
                    self.set_resizable(False)

                # open the app with a minimal UI when opening a single file
                self.left_button.set_visible(False)
                self.menu_button.set_visible(False)

            return True

        elif len(files) > 1 and self.multi_install.set_from_local_files(files):
            self.container_stack.set_visible_child(self.multi_install)

            if self.from_file:
                # open the app with a minimal UI when opening a single file
                self.left_button.set_visible(False)
                self.menu_button.set_visible(False)

            return True

        utils.send_notification(
            Gio.Notification.new('Unsupported file type: Gear lever can\'t handle these types of files.')
        )

        return False

    def on_multi_install_show_details(self, source: MultiInstall, el: AppImageListElement):
        file = Gio.File.new_for_path(el.file_path)
        self.on_selected_local_file([file])

    def on_show_installed_list(self, source: Gtk.Widget=None, data=None):
        self.container_stack.set_transition_type(Adw.LeafletTransitionType.OVER)

        # self.installed_apps_list.refresh_list()
        self.container_stack.set_visible_child(self.app_lists_stack)

    def on_left_button_clicked(self, *args):
        if self.app_lists_stack.get_visible_child() == self.installed_stack:
            container_visible = self.container_stack.get_visible_child()

            if container_visible == self.app_details:
                if self.selected_files_count > 1:
                    self.container_stack.set_visible_child(self.multi_install)

                else:
                    self.titlebar.set_title_widget(self.view_title_widget)
                    self.on_show_installed_list()

            elif container_visible == self.multi_install:
                self.titlebar.set_title_widget(self.view_title_widget)
                self.on_show_installed_list()

            elif container_visible == self.multi_update:
                self.titlebar.set_title_widget(self.view_title_widget)
                self.on_show_installed_list()

            elif container_visible == self.app_lists_stack:
                self.on_open_file_chooser()

    def on_app_lists_stack_change(self, widget, data):
        pass

    def on_update_all_event(self, *args):
        self.container_stack.set_visible_child(self.multi_update)
        self.multi_update.start()

    def on_container_stack_change(self, widget, data):
        in_app_details = self.container_stack.get_visible_child() is self.app_details
        in_multi_install = self.container_stack.get_visible_child() is self.multi_install
        in_multi_update = self.container_stack.get_visible_child() is self.multi_update
        in_apps_list = self.container_stack.get_visible_child() is self.app_lists_stack

        if in_app_details or in_multi_install:
            self.left_button.set_icon_name('gl-left-symbolic')
            self.left_button.set_tooltip_text(None)
            self.search_btn.set_visible(False)
        elif in_multi_update:
            self.left_button.set_visible(False)
        else:
            self.search_btn.set_visible(True)
            self.left_button.set_child(self.open_appimage_button_child)

        if in_apps_list:
            self.installed_apps_list.refresh_list()

        self.view_title_widget.set_visible(not in_app_details)

    def on_drop_event(self, widget, value, x, y):
        if isinstance(value, Gdk.FileList):
            logging.debug('Opening file from drag and drop')
            self.selected_files_count = len(list(value))
            return self.on_selected_local_file(list(value))

        return False

    def on_drop_enter(self, widget, x, y):
        self.container_stack.set_transition_type(Adw.LeafletTransitionType.UNDER)
        self.visible_before_dragdrop_start = self.container_stack.get_visible_child()
        self.container_stack.set_visible_child(self.drag_drop_ui)

        return Gdk.DragAction.COPY
    
    def on_drop_leave(self, widget):
        self.container_stack.set_transition_type(Adw.LeafletTransitionType.UNDER)

        if self.visible_before_dragdrop_start:
            self.container_stack.set_visible_child(self.visible_before_dragdrop_start)
        else:
            self.container_stack.set_visible_child(self.app_lists_stack)

        return Gdk.DragAction.COPY
    
    def on_uninstalled_app(self, widget, data):
        if self.from_file:
            return self.close()

        self.on_show_installed_list(widget, data)

    def on_app_update_started(self, *args):
        self.left_button.set_visible(False)
        self.left_button.set_sensitive(False)

    def on_app_update_ended(self, *args):
        self.left_button.set_visible(True)
        self.left_button.set_sensitive(True)


    def on_open_file_chooser_response(self, dialog, result):
        try:
            selected_files = dialog.open_multiple_finish(result)
            self.selected_files_count = len(list(selected_files))
        except Exception as e:
            logging.error(str(e))
            return

        if selected_files:
            self.on_selected_local_file(list(selected_files))

    def on_open_file_chooser(self):
        dialog = Gtk.FileDialog(title=_('Open a file'), modal=True)

        dialog.open_multiple(
            parent=self,
            cancellable=None,
            callback=self.on_open_file_chooser_response
        )

    def on_close_request(self, widget):
        appimage_provider.extraction_folder_cleanup()

    def on_window_maximixed_changed(self, *args):
        self.settings.set_boolean('is-maximized', self.is_maximized())

    def on_trigger_search_mode(self, *args):
        self.installed_apps_list.trigger_search_mode()



import gi
import os
import configparser
import hashlib
import logging

gi.require_version('Gtk', '4.0')
gi.require_version('Adw', '1')

from gi.repository import GLib  # noqa


class Config:
    path = os.path.join(GLib.get_user_config_dir(), 'gearlever.conf')
    parser = configparser.ConfigParser()

    @staticmethod
    def return_boolean(v):
        return 'yes' if v else 'no'

    @staticmethod
    def exists():
        return os.path.exists(Config.path)

    @staticmethod
    def refresh():
        if Config.exists():
            Config.parser.read(Config.path)

    @staticmethod
    def write():
        logging.info(f'Writing config to {Config.path}')
        with open(Config.path, 'w') as f:
            Config.parser.write(f)

    @staticmethod
    def get_app_hash(el):
        return hashlib.md5(el.file_path.encode()).hexdigest()

    @staticmethod
    def get_app_config(el):
        h = Config.get_app_hash(el)
        k = f'app.{h}'

        if Config.parser.has_section(k):
            return dict(Config.parser[f'app.{h}'])
        return {}

    @staticmethod
    def delete_app_config(el):
        h = Config.get_app_hash(el)
        j = f'app.{h}'
        k = f'app.{h}.update_manager'

        for l in [j, k]:
            if Config.parser.has_section(l):
                logging.info(f'Deleting config section {l}')
                Config.parser.remove_section(l)

        Config.write()

    @staticmethod
    def set_app_config(el, data: dict):
        h = Config.get_app_hash(el)
        logging.info(f'Setting app config for {el.name} (app.{h}): {data}')
        data['name'] = el.name
        data['file_path'] = el.file_path
        Config.parser[f'app.{h}'] = data
        Config.write()

    @staticmethod
    def get_app_update_config(el):
        h = Config.get_app_hash(el)
        k = f'app.{h}.update_manager'

        if not Config.parser.has_section(k):
            Config.parser.add_section(k)

        return Config.parser[f'app.{h}.update_manager']

    @staticmethod
    def set_app_update_config(el, manager, data):
        h = Config.get_app_hash(el)
        logging.info(f'Setting update config for {el.name} (app.{h}.update_manager), manager={manager.name}: {data}')
        Config.parser[f'app.{h}.update_manager'] = data
        Config.parser[f'app.{h}.update_manager']['manager'] = manager.name
        Config.write()

    @staticmethod
    def delete_app_update_config(el):
        h = Config.get_app_hash(el)
        k = f'app.{h}.update_manager'
        if Config.parser.has_section(k):
            logging.info(f'Deleting update config section {k}')
            Config.parser.remove_section(k)

        Config.write()


import logging

from .AppListElement import AppListElement
from typing import Optional

class AppUpdateElement():
    def __init__(self, app_id: str, size: Optional[str], to_verison: Optional[str], **kwargs):
        self.id: str = app_id
        self.size: Optional[str] = size
        self.to_version: Optional[str] = to_verison
        self.extra_data: dict = {}

        for k, v in kwargs.items():
            self.extra_data[k] = v

class InternalError(Exception):
    def __init__(self, message: str, *args) -> None:
        super().__init__(*args)
        self.message = message

        logging.error(message)

class DownloadInterruptedException(Exception):
    pass


import logging
from gi.repository import Gtk, GObject, Adw, Gdk, Gio, GLib

from .models.AppListElement import InstalledStatus
from .providers.AppImageProvider import AppImageListElement, AppImageUpdateLogic
from .providers.providers_list import appimage_provider
from .lib.async_utils import _async, idle, debounce
from .lib.utils import get_application_window
from .models.UpdateManagerChecker import UpdateManagerChecker
from .components.AppListBoxItem import AppListBoxItem

class MultiUpdate(Gtk.ScrolledWindow):
    __gsignals__ = {
        "go-back": (GObject.SIGNAL_RUN_FIRST, GObject.TYPE_NONE, (bool, )),
    }

    def __init__(self) -> None:
        super().__init__()
        self.ACTION_ROW_ICON_SIZE = 45

        self.app_list: list[AppImageListElement] = []
        self.app_list_box_items: list[AppListBoxItem] = []
        self.app_list_box = Gtk.ListBox(css_classes=['boxed-list'])
        self.current_update_manager = None
        self.green_light = True

        self.cancel_update_btn = Gtk.Button(
            label=_('Cancel update'), halign=Gtk.Align.CENTER)
        
        view_box = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL, 
            spacing=10
        )

        main_box = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL, 
            margin_top=30, 
            margin_bottom=10, 
            margin_start=20, 
            margin_end=20,
            spacing=20
        )

        main_box.append(self.cancel_update_btn)
        main_box.append(self.app_list_box)

        clamp = Adw.Clamp(child=main_box)

        self.progress_bar = Gtk.ProgressBar(
            css_classes=['osd', 'horizontal', 'top'], fraction=1
        )

        view_box.append(self.progress_bar)
        view_box.append(clamp)

        self.cancel_update_btn.connect('clicked', self.on_cancel_update_btn)
        self.set_child(view_box)

    @idle
    def update_progress_fraction(self, p):
        self.progress_bar.set_fraction(p)

    def count_not_installed(self):
        not_installed_count = 0
        for el in self.app_list:
            if el.installed_status is InstalledStatus.NOT_INSTALLED:
                not_installed_count += 1

        return not_installed_count

    @idle
    def create_app_row(self, el: AppImageListElement):
        appimage_provider.refresh_data(el)
        icon = appimage_provider.get_icon(el)

        row = AppListBoxItem(el, show_details_btn=False)
        row.set_icon(icon)
        row.set_opacity(0.5)

        self.app_list_box.append(row)
        self.app_list_box_items.append(row)

    @idle
    def on_update_end(self):
        self.emit('go-back', None)

    @idle
    def mark_as_updated(self, el: AppImageListElement):
        for row in self.app_list_box_items:
            if row._app.file_path == el.file_path:
                row.set_opacity(1)
                return

    @_async
    def update_all(self):
        p = len(self.app_list)
        for i, el in enumerate(self.app_list):
            if not self.green_light:
                break

            manager = UpdateManagerChecker.check_url_for_app(el)
            self.current_update_manager = manager
            appimage_provider.update_from_url(manager, el, 
                status_cb=lambda s: self.update_progress_fraction((s / p) * (i + 1)))
            self.mark_as_updated(el)

        self.update_progress_fraction(1)
        self.on_update_end()

    def start(self):
        self.green_light = True
        self.progress_bar.set_fraction(0)
        self.app_list_box.remove_all()
        self.app_list = []
        self.app_list_box_items = []
        self.check_updatables()

    @_async
    def check_updatables(self):
        installed = appimage_provider.list_installed()

        for el in installed:
            manager = UpdateManagerChecker.check_url_for_app(el)
            if not manager:
                continue

            try:
                if manager.is_update_available() and \
                    (not appimage_provider.is_app_running(el)):
                    self.app_list.append(el)
                    self.create_app_row(el)
            except Exception as e:
                pass

        self.update_all()

    def on_cancel_update_btn(self, button):
        self.green_light = False
        if self.current_update_manager:
            self.current_update_manager.cancel_download()


import time
import logging
import base64
import os
import traceback
import platform
import shlex
from typing import Optional, Callable
from gi.repository import Gtk, GObject, Adw, Gdk, Gio, Pango, GLib

from .lib.ini_config import Config
from .lib.terminal import sandbox_sh
from .models.Models import InternalError
from .models.UpdateManager import UpdateManager
from .models.UpdateManagerChecker import UpdateManagerChecker
from .models.AppListElement import InstalledStatus
from .providers.AppImageProvider import AppImageListElement, AppImageUpdateLogic
from .providers.providers_list import appimage_provider
from .lib.async_utils import _async, idle, debounce
from .lib.utils import url_is_valid, get_file_hash, get_application_window, show_message_dialog, gnu_naturalsize, check_internet
from .components.CustomComponents import CenteringBox, LabelStart
from .components.AppDetailsConflictModal import AppDetailsConflictModal
from .components.AdwEntryRowDefault import AdwEntryRowDefault


class AppDetails(Gtk.ScrolledWindow):
    """The presentation screen for an application"""
    __gsignals__ = {
        "uninstalled-app": (GObject.SIGNAL_RUN_FIRST, GObject.TYPE_NONE, (object, )),
        "update-started": (GObject.SIGNAL_RUN_FIRST, GObject.TYPE_NONE, (object, )),
        "update-ended": (GObject.SIGNAL_RUN_FIRST, GObject.TYPE_NONE, (object, )),
    }

    UPDATE_BTN_LABEL = _('Update')
    BTN_SAVE = _('Save')
    BTN_SAVED = _('Saved')
    UPDATE_BTN_ERROR = _('Invalid data')
    CANCEL_UPDATE = _('Cancel update')
    UPDATE_FETCHING = _('Checking updates...')
    UPDATE_NOT_AVAIL_BTN_LABEL = _('No updates available')
    UPDATE_NO_MATCH_BTN_LABEL = _('Release file not found')
    UPDATE_INFO_EMBEDDED = _('This application includes update information provided by the developer')
    UPDATE_INFO_NOT_EMBEDDED = _('Manage update details for this application')


    def __init__(self):
        super().__init__()
        self.ACTION_ROW_ICON_SIZE = 34
        self.EXTRA_DATA_SPACING = 20
        self.minimal_ui = False

        self.app_list_element: AppImageListElement | None = None
        self.common_btn_css_classes = ['big-btn', 'text-button']

        self.main_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, margin_top=10, margin_bottom=10, margin_start=20, margin_end=20)

        # 1st row
        self.details_row = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
        self.icon_slot = Gtk.Box()

        title_col = CenteringBox(orientation=Gtk.Orientation.VERTICAL, hexpand=True, spacing=2, margin_top=5)
        self.title = Gtk.Label(label='', css_classes=['title-1'], hexpand=True, halign=Gtk.Align.CENTER, wrap=True)
        self.app_subtitle = Gtk.Label(
            label='',
            halign=Gtk.Align.CENTER,
            css_classes=['dim-label', 'subtitle'],
            ellipsize=Pango.EllipsizeMode.END,
            max_width_chars=50,
            lines=2,
            justify=Gtk.Justification.CENTER,
            selectable=True,
        )

        self.app_info_pills = Adw.ToggleGroup(can_target=False, css_classes=['round'], 
                                              halign=Gtk.Align.CENTER,
                                              margin_top=20,
                                              margin_bottom=10)

        [title_col.append(el) for el in [self.title, self.app_subtitle]]

        self.source_selector_hdlr = None
        self.source_selector = Gtk.ComboBoxText()
        self.source_selector_revealer = Gtk.Revealer(child=self.source_selector, transition_type=Gtk.RevealerTransitionType.CROSSFADE)

        # Action buttons
        self.primary_action_button = Gtk.Button(label='', valign=Gtk.Align.CENTER, halign=Gtk.Align.CENTER, 
                                                css_classes=self.common_btn_css_classes, width_request=200)
        self.secondary_action_button = Gtk.Button(label='', valign=Gtk.Align.CENTER, 
                                            css_classes=self.common_btn_css_classes, width_request=200)
        self.update_action_button = Gtk.Button(
            label=_('Update'), 
            valign=Gtk.Align.CENTER, 
            width_request=200,
            css_classes=[*self.common_btn_css_classes, 'suggested-action'],
            visible=False
        )

        action_buttons_row = CenteringBox(orientation=Gtk.Orientation.VERTICAL, spacing=10, margin_top=20)
        self.primary_action_button.connect('clicked', self.on_primary_action_button_clicked)
        self.secondary_action_button.connect('clicked', self.on_secondary_action_button_clicked)
        self.update_action_button.connect('clicked', self.update_action_button_clicked)
        
        self.primary_action_buttons_row = CenteringBox(orientation=Gtk.Orientation.HORIZONTAL, spacing=10)
        [self.primary_action_buttons_row.append(el) for el in [self.secondary_action_button, self.primary_action_button]]
        [action_buttons_row.append(el) for el in [self.primary_action_buttons_row, self.update_action_button]]

        [self.details_row.append(el) for el in [self.icon_slot, title_col, self.app_info_pills, action_buttons_row]]

        # row
        self.spinner_row = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, margin_top=20, margin_bottom=20)

        self.desc_row_spinner = Gtk.Spinner(spinning=True, visible=True)
        [self.spinner_row.append(el) for el in [self.desc_row_spinner]]

        # row
        self.third_row = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.extra_data = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=self.EXTRA_DATA_SPACING)
        self.third_row.append(self.extra_data)

        [self.main_box.append(el) for el in [self.details_row, self.spinner_row, self.third_row]]

        clamp = Adw.Clamp(child=self.main_box, maximum_size=600, margin_top=10, margin_bottom=20)

        # Window top banner
        self.window_banner = Adw.Banner(use_markup=True)
        self.window_banner.connect('button-clicked', self.after_trust_buttons_interaction)

        container_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        [container_box.append(el) for el in [self.window_banner, clamp]]

        self.env_variables_widgets = []
        self.env_variables_group_container = None
        self.save_vars_btn: Optional[Gtk.Button] = None
        self.command_line_arguments_row = None

        # Update url entry
        self.update_manager: Optional[UpdateManager] = None
        self.update_url_group: Optional[Adw.PreferencesGroup] = None
        self.update_url_row: Optional[AdwEntryRowDefault] = None
        self.update_url_save_btn: Optional[Gtk.Button] = None
        self.update_url_source: Optional[Adw.ComboRow] = None
        self.update_url_form_rows: list = []
        self.update_url_group: Optional[Adw.PreferenciesGroup] = None

        self.set_child(container_box)

    def set_app_list_element(self, el: AppImageListElement):
        self.app_list_element = el
        self.provider = appimage_provider
        self.update_action_button.set_visible(False)

        self.before_load()
        self.load()

    def set_from_local_file(self, file: Gio.File):
        if appimage_provider.can_install_file(file):
            list_element = appimage_provider.create_list_element_from_file(file)
            self.set_app_list_element(list_element)
            return True

        logging.warn('Trying to open an unsupported file')
        return False

    @idle
    def complete_load(self, icon: Gtk.Image, generation: str, load_completed_callback: Optional[Callable] = None):
        self.show_row_spinner(True)

        if self.icon_slot:
            self.details_row.remove(self.icon_slot)

        self.icon_slot = icon
        icon.set_pixel_size(128)

        self.details_row.prepend(self.icon_slot)
        self.title.set_label(self.app_list_element.name)

        self.app_subtitle.set_text(self.app_list_element.description)
        self.app_subtitle.set_visible(
            len(self.app_list_element.description.strip()) > 0
        )

        self.app_info_pills.remove_all()

        if self.app_list_element.version:
            t1 = Adw.Toggle(label=self.app_list_element.version)
            self.app_info_pills.add(t1)

        t2 = Adw.Toggle(label=gnu_naturalsize(self.app_list_element.size))
        self.app_info_pills.add(t2)

        if generation:
            t3 = Adw.Toggle(label=f'{self.provider.name} Type. {generation}')
            self.app_info_pills.add(t3)

        self.app_info_pills.set_active(0)
        self.app_info_pills.set_visible(not self.minimal_ui)

        if self.extra_data:
            self.third_row.remove(self.extra_data)

        self.extra_data = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=self.EXTRA_DATA_SPACING)
        self.third_row.append(self.extra_data)

        self.install_button_label_info = None

        # Load the boxed list with additional information
        gtk_list = Gtk.ListBox(css_classes=['boxed-list'])

        # The path of the executable
        gtk_list.append(self.create_exec_path_row())

        # Hashes
        if self.app_list_element.installed_status is not InstalledStatus.INSTALLED:
            gtk_list.append(self.create_app_hash_row())
        
        self.update_installation_status()

        self.secondary_action_button.set_visible(True)
        self.primary_action_button.set_visible(True)
        self.title.set_visible(True)

        system_arch = platform.machine()

        if self.app_list_element.installed_status is InstalledStatus.INSTALLED:
            # Exec arguments
            self.command_line_arguments_row = self.create_show_exec_args_row()
            gtk_list.append(self.command_line_arguments_row)

            self.primary_action_buttons_row.set_orientation(Gtk.Orientation.HORIZONTAL)

            # A custom link to a website
            gtk_list.append(self.create_edit_custom_website_row())

            # Reload metadata row
            reload_data_listbox = Gtk.ListBox(css_classes=['boxed-list'])
            reload_data_listbox.append(self.create_reload_metadata_row())
            self.extra_data.prepend(reload_data_listbox)

            # Show or hide window banner
            if system_arch and system_arch != self.app_list_element.architecture:
                self.show_invalid_arch_banner()
            elif self.app_list_element.external_folder:
                self.window_banner.set_revealed(True)
                self.window_banner.set_button_label(None)
                self.window_banner.set_title(_('This app is located outside the default folder\n<small>You can hide external apps in the settings</small>'))
            else:
                self.window_banner.set_revealed(False)
        else:
            self.primary_action_buttons_row.set_orientation(Gtk.Orientation.VERTICAL)

            if self.app_list_element.trusted:
                if system_arch and system_arch != self.app_list_element.architecture:
                    self.show_invalid_arch_banner()
                else:
                    self.window_banner.set_revealed(False)
            else:
                self.window_banner.set_revealed(True)
                self.window_banner.set_title(_('Please, verify the source of this app before opening it'))
                self.window_banner.set_button_label(_('Unlock'))
                self.secondary_action_button.set_sensitive(False)
                self.primary_action_button.set_sensitive(False)

        self.extra_data.append(gtk_list)

        if self.app_list_element.installed_status is InstalledStatus.INSTALLED:
            self.update_url_group = self.create_edit_update_url_row()
            self.extra_data.append(self.update_url_group)

            edit_env_vars_widget = self.create_edit_env_vars_row()
            self.extra_data.append(edit_env_vars_widget)
            self.update_action_button.set_visible(False)

        self.show_row_spinner(False)

        if load_completed_callback:
            load_completed_callback()

    @_async
    def load(self, load_completed_callback: Optional[Callable] = None):
        if self.command_line_arguments_row:
            self.command_line_arguments_row.remove_css_class('error')

        icon = Gtk.Image(icon_name='application-x-executable-symbolic')
        generation = self.provider.get_appimage_type(self.app_list_element)

        try:
            if self.app_list_element.trusted:
                icon = self.provider.get_icon(self.app_list_element)
                self.provider.refresh_arch(self.app_list_element)

                if self.app_list_element.installed_status is not InstalledStatus.INSTALLED:
                    self.provider.refresh_data(self.app_list_element)

            self.complete_load(
                icon,
                generation,
                load_completed_callback=load_completed_callback
            )
        except Exception as e:
            self.handle_exception(e)
            self.emit('uninstalled-app', self)

    def before_load(self):
        self.show_row_spinner(True)
        self.title.set_visible(False)
        self.app_subtitle.set_visible(False)
        self.secondary_action_button.set_visible(False)
        self.primary_action_button.set_visible(False)
        self.window_banner.set_revealed(False)
        self.app_info_pills.set_visible(False)

        if self.icon_slot:
            self.details_row.remove(self.icon_slot)
            self.icon_slot = None

        if self.extra_data:
            self.third_row.remove(self.extra_data)
            self.extra_data = None

    def set_minimal_ui(self, enabled=True):
        self.minimal_ui = enabled
        self.third_row.set_visible(not enabled)
        self.app_info_pills.set_visible(not enabled)

        if enabled:
            self.primary_action_buttons_row.set_orientation(Gtk.Orientation.VERTICAL)
            self.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.NEVER)
        else:
            self.primary_action_buttons_row.set_orientation(Gtk.Orientation.HORIZONTAL)
            self.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)

    @_async
    def install_file(self, el: AppImageListElement):
        try:
            self.provider.install_file(el)
            self.update_installation_status()

            self.complete_load(
                self.provider.get_icon(self.app_list_element),
                self.provider.get_appimage_type(self.app_list_element),
            )
        except Exception as e:
            self.handle_exception(e)
            self.emit('uninstalled-app', self)
            
    def handle_exception(self, e: Exception):
        if type(e) is InternalError:
                show_message_dialog(
                    e.message,
                    markup=True
                )
        else:
            traceback.print_exc()
            show_message_dialog(
                '',
                header=_('Unkown Error'),
                markup=False
            )

    def on_conflict_modal_close(self, widget, data: str):
        if data == 'cancel':
            self.app_list_element.update_logic = None
            return

        self.app_list_element.update_logic = AppImageUpdateLogic[data]
        self.on_primary_action_button_clicked()

    def on_primary_action_button_clicked(self, button: Optional[Gtk.Button] = None):
        if self.app_list_element.installed_status == InstalledStatus.INSTALLED:
            self.show_remove_confirm_dialog()
        elif self.app_list_element.installed_status == InstalledStatus.NOT_INSTALLED:
            if self.provider.is_updatable(self.app_list_element) and not self.app_list_element.update_logic:
                confirm_modal = AppDetailsConflictModal(app_name=self.app_list_element.name)

                confirm_modal.modal.connect('response', self.on_conflict_modal_close)
                confirm_modal.modal.present()
                return

            self.app_list_element.set_installed_status(InstalledStatus.INSTALLING)

            if self.app_list_element.trusted:
                self.update_installation_status()

                if self.app_list_element.update_logic and (self.app_list_element.update_logic == AppImageUpdateLogic.REPLACE):
                    old_version = next(
                        filter(lambda old_v: (old_v.name == self.app_list_element.name), self.provider.list_installed()), 
                        None
                    )

                    self.app_list_element.updating_from = old_version
                    self.provider.uninstall(old_version)

                self.install_file(self.app_list_element)

        self.update_installation_status()

    def on_secondary_action_button_clicked(self, button: Gtk.Button):
        if self.app_list_element.installed_status in [InstalledStatus.INSTALLED, InstalledStatus.NOT_INSTALLED]:
            is_terminal = self.app_list_element.desktop_entry and self.app_list_element.desktop_entry.getTerminal()
            if self.app_list_element.trusted and (not is_terminal):
                try:
                    self.app_list_element.set_trusted()
                    
                    pre_launch_label = self.secondary_action_button.get_label()
                    self.secondary_action_button.set_label(_('Launching...'))
                    self.secondary_action_button.set_sensitive(False)

                    try:
                        self.provider.run(self.app_list_element)
                    except Exception as e:
                        traceback.print_exc()
                        self.handle_exception(e)

                    self.post_launch_animation(restore_as=pre_launch_label)

                except Exception as e:
                    logging.error(str(e))
        elif self.app_list_element.installed_status == InstalledStatus.UPDATING:
            if self.update_manager:
                self.update_manager.cancel_download()
                self.update_installation_status()

    def on_remove_app_clicked(self, dialog, response: str):
        if response == 'remove':
            self.app_list_element.set_installed_status(InstalledStatus.UNINSTALLING)
            self.update_installation_status()

            self.provider.uninstall(self.app_list_element)
            self.emit('uninstalled-app', self)

    @_async
    def post_launch_animation(self, restore_as):
        GLib.idle_add(lambda: self.secondary_action_button.set_sensitive(False))
        time.sleep(3)

        GLib.idle_add(lambda: self.secondary_action_button.set_label(restore_as))
        GLib.idle_add(lambda: self.secondary_action_button.set_sensitive(True))

    @_async
    def update_action_button_clicked(self, w):
        is_running = self.provider.is_app_running(self.app_list_element)

        if is_running:
            self.show_update_error_dialog(
                _('{app_name} is running, please close it before proceeding with the update'.format(
                    app_name=self.app_list_element.name))
            )

            return

        self.app_list_element.set_installed_status(InstalledStatus.UPDATING)
        self.update_installation_status()

        if not self.update_manager:
            return

        update_success = False

        try:
            self.emit('update-started', None)
            update_result = appimage_provider.update_from_url(self.update_manager, self.app_list_element, status_cb= lambda s: \
                GLib.idle_add(lambda:  self.update_action_button.set_label(str(round(s * 100)) + ' %')
            ))

            if update_result == None:
                update_success = False
            else:
                update_success = True
                self.app_list_element = update_result

        except Exception as e:
            logging.error('Update error', exc_info=e)
            self.show_update_error_dialog(str(e))
        finally:
            self.emit('update-ended', None)

        GLib.idle_add(lambda: self.update_action_button.set_label(_('Update')))
        GLib.idle_add(lambda: self.update_action_button.set_sensitive(True))
        self.app_list_element.set_installed_status(InstalledStatus.INSTALLED)
        self.update_manager.cleanup()

        if update_success:
            self.update_manager = None

            icon = self.provider.get_icon(self.app_list_element)
            self.provider.refresh_data(self.app_list_element)

            generation = self.provider.get_appimage_type(self.app_list_element)
            self.complete_load(icon, generation)
            self.update_installation_status()


    @idle
    def show_update_error_dialog(self, msg: str):
        logging.error(msg)
        dialog = Adw.MessageDialog(
            transient_for=get_application_window(),
            heading=_('Update error'),
            body=msg
        )

        dialog.add_response('okay', _('Close'))
        dialog.present()

    @idle
    def show_remove_confirm_dialog(self):
        dialog = Adw.MessageDialog(
            transient_for=get_application_window(),
            heading=_('Do you really want to remove this app?'),
        )

        dialog.add_response('cancel', _('Cancel'))
        dialog.add_response('remove', _('Remove'))
        dialog.set_response_appearance('remove', Adw.ResponseAppearance.DESTRUCTIVE)
        dialog.set_close_response('cancel')
        dialog.set_close_response('remove')
        dialog.connect('response', self.on_remove_app_clicked)

        dialog.present()

    @idle
    def set_all_btn_sensitivity(self, s: bool):
        self.primary_action_button.set_sensitive(s)
        self.secondary_action_button.set_sensitive(s)
        self.update_action_button.set_sensitive(s)

    @idle
    def restore_launch_button(self, restore_as):
        self.secondary_action_button.set_label(restore_as)
        self.secondary_action_button.set_sensitive(True)

    def update_status_callback(self, status: bool):
        if not status:
            self.app_list_element.set_installed_status(InstalledStatus.ERROR)

        self.update_installation_status()

    def update_installation_status(self):
        self.primary_action_button.set_css_classes(self.common_btn_css_classes)
        self.secondary_action_button.set_css_classes(self.common_btn_css_classes)
        self.source_selector.set_visible(False)

        self.primary_action_button.set_sensitive(True)
        self.secondary_action_button.set_sensitive(True)

        if self.app_list_element.installed_status == InstalledStatus.INSTALLED:
            if self.app_list_element.desktop_entry and self.app_list_element.desktop_entry.getTerminal():
                self.secondary_action_button.set_label(_('This app runs in the terminal'))
            else:
                self.secondary_action_button.set_label(_('Launch'))

            self.primary_action_button.set_label(_('Remove'))
            self.primary_action_button.set_css_classes([*self.common_btn_css_classes, 'destructive-action'])

        elif self.app_list_element.installed_status == InstalledStatus.UNINSTALLING:
            self.primary_action_button.set_label(_('Uninstalling...'))
            self.primary_action_button.set_sensitive(False)

        elif self.app_list_element.installed_status == InstalledStatus.INSTALLING:
            self.primary_action_button.set_label(_('Installing...'))
            self.primary_action_button.set_sensitive(False)

        elif self.app_list_element.installed_status == InstalledStatus.NOT_INSTALLED:
            if not self.app_list_element.desktop_entry:
                self.secondary_action_button.set_label(_('Launch'))
            else:
                if self.app_list_element.desktop_entry.getTerminal():
                    self.secondary_action_button.set_label(_('This app runs in the terminal'))
                    self.secondary_action_button.set_sensitive(False)
                else:
                    self.secondary_action_button.set_label(_('Launch'))

            self.primary_action_button.set_label(_('Move to the app menu'))
            self.primary_action_button.set_css_classes([*self.common_btn_css_classes, 'suggested-action'])

        elif self.app_list_element.installed_status == InstalledStatus.UPDATE_AVAILABLE:
            pass
        elif self.app_list_element.installed_status == InstalledStatus.UPDATING:
            self.primary_action_button.set_sensitive(False)
            self.secondary_action_button.set_label(self.CANCEL_UPDATE)
            self.secondary_action_button.set_sensitive(True)
            self.secondary_action_button.set_css_classes([*self.common_btn_css_classes])
            self.update_action_button.set_sensitive(False)

        elif self.app_list_element.installed_status == InstalledStatus.ERROR:
            self.primary_action_button.set_label(_('Error'))
            self.primary_action_button.set_css_classes([*self.common_btn_css_classes, 'destructive-action'])

    def provider_refresh_installed_status(self, status: Optional[InstalledStatus] = None):
        if status:
            self.app_list_element.installed_status = status

        self.update_installation_status()

    def show_invalid_arch_banner(self):
        self.window_banner.set_revealed(True)
        self.window_banner.set_title(_('This app might not be compatible with your system architecture'))
        self.window_banner.set_button_label('')

    @idle
    def show_row_spinner(self, status: bool):
        self.desc_row_spinner.set_visible(status)
        self.desc_row_spinner.set_spinning(status)

    def after_trust_buttons_interaction(self, widget: Adw.Banner):
        if not self.app_list_element:
            return
        
        self.app_list_element.trusted = True
        widget.set_revealed(False)

        self.before_load()
        self.load()

    @debounce(0.5)
    @idle
    def on_web_browser_input_apply(self, widget):
        app_conf = self.get_config_for_app()

        text = widget.get_text().strip()

        widget.remove_css_class('error')
        if text and (not url_is_valid(text)):
            return widget.add_css_class('error')

        if text:
            widget.add_css_class('success')
        else:
            widget.remove_css_class('success')
            widget.remove_css_class('error')

        app_conf['website'] = text
        Config.set_app_config(self.app_list_element, app_conf)

    @idle
    def set_app_as_updatable(self):
        self.primary_action_buttons_row.set_orientation(Gtk.Orientation.HORIZONTAL)
        self.update_action_button.set_visible(True)
        self.update_action_button.set_label(self.UPDATE_FETCHING)
        self.update_action_button.set_sensitive(False)

    @_async
    def check_updates(self):
        if not check_internet():
            return

        manager = self.update_manager

        if not manager:
            GLib.idle_add(lambda: self.update_url_save_btn.set_visible(True))
            return

        self.set_app_as_updatable()

        is_updatable = False

        try:
            is_updatable = manager.is_update_available()
        except Exception as e:
            logging.error('Update error', exc_info=e)
            self.show_update_error_dialog(str(e))

        self.app_list_element.is_updatable_from_url = is_updatable

        if is_updatable:
            logging.debug(f'{self.app_list_element.name} is_updatable')
            GLib.idle_add(lambda: self.update_action_button.set_label(self.UPDATE_BTN_LABEL))
        elif is_updatable is None:
            GLib.idle_add(lambda: self.update_action_button.set_label(self.UPDATE_NO_MATCH_BTN_LABEL))
        else:
            GLib.idle_add(lambda: self.update_action_button.set_label(self.UPDATE_NOT_AVAIL_BTN_LABEL))

        GLib.idle_add(lambda: self.update_action_button.set_sensitive(True))
        GLib.idle_add(lambda: self.update_action_button.set_sensitive(is_updatable))

    @idle
    def set_update_information(self, manager: Optional[UpdateManager]=None):
        if manager:
            self.update_manager = manager
        else:
            manager_label = self.update_url_source.get_model().get_string(
                self.update_url_source.get_selected())
            
            selected_model = None
            for m in UpdateManagerChecker.get_models():
                if m.label == manager_label:
                    selected_model = m
                    break

            if selected_model:
                self.update_manager = selected_model(el=self.app_list_element, embedded=False)
            else:
                if self.app_list_element:
                    Config.delete_app_update_config(self.app_list_element)

                self.update_manager = UpdateManagerChecker.check_url_for_app(el=self.app_list_element)

        for r in self.update_url_form_rows:
            self.update_url_group.remove(r)   

        self.update_url_form_rows = []

        if self.update_manager:
            rows = self.update_manager.load_form_rows()

            for r in rows:
                if isinstance(r, Adw.SwitchRow):
                    r.connect('notify::active', self.on_update_form_rows_change)
                elif isinstance(r, Adw.EntryRow):
                    r.connect('changed', self.on_update_form_rows_change)
                    
                self.update_url_form_rows.append(r)
                self.update_url_group.add(r)

            if self.update_manager.embedded:
                desc = self.UPDATE_INFO_EMBEDDED + f' ({self.update_manager.label})'
                self.update_url_group.set_description(desc)
            else:
                self.update_url_group.set_description(self.UPDATE_INFO_NOT_EMBEDDED)

            self.update_url_save_btn.set_sensitive(False)
            self.check_updates()

    def on_update_form_rows_change(self, *args):
        self.update_url_save_btn.set_sensitive(True)

    @idle
    def on_app_update_url_success(self):
        self.update_url_save_btn.set_sensitive(False)
        self.update_url_save_btn.add_css_class('success')
        self.update_url_save_btn.set_label(self.BTN_SAVED)

    @idle
    def on_app_update_url_error(self):
        self.update_url_save_btn.set_sensitive(False)
        self.update_url_save_btn.add_css_class('error')
        self.update_url_save_btn.set_label(self.UPDATE_BTN_ERROR)

    @idle
    def on_app_update_url_reset(self):
        self.update_url_save_btn.set_sensitive(True)
        self.update_url_save_btn.remove_css_class('error')
        self.update_url_save_btn.remove_css_class('success')
        self.update_url_save_btn.set_child(
            Adw.ButtonContent(
            icon_name='gearlever-check-plain-symbolic',
            label=_(self.BTN_SAVE),
        ))

    @_async
    def on_app_update_url_apply(self, button):
        app_conf = self.get_config_for_app()
        def_sleep = 2

        self.on_app_update_url_reset()

        if not self.update_manager:
            if self.app_list_element:
                Config.delete_app_update_config(self.app_list_element)

            self.on_app_update_url_success()
            return

        # text = self.update_manager.get_url_from_form()
        # text = text.strip()

        # if not text:
        #     self.on_app_update_url_error()

        #     time.sleep(def_sleep)
        #     self.on_app_update_url_reset()
        #     return

        # if not self.update_manager.can_handle_link(url=text):
        #     self.on_app_update_url_error()
            
        #     time.sleep(def_sleep)
        #     self.on_app_update_url_reset()
        #     return

        # self.update_manager.set_url(text)
        try:
            form_config = self.update_manager.get_config_from_form()
            self.update_manager.validate_config(form_config)
        except Exception as e:
            logging.error(traceback.format_exc())
            self.show_update_error_dialog(str(e))

            time.sleep(def_sleep)
            self.on_app_update_url_reset()
            return

        if not self.update_manager.embedded:
            Config.set_app_update_config(self.app_list_element, self.update_manager, form_config)

        self.on_app_update_url_success()

        self.check_updates()

        time.sleep(def_sleep)
        self.on_app_update_url_reset()

    def on_env_var_value_changed(self, widget, key_widget, value_widget):
        key = key_widget.get_text()
        value_widget.set_sensitive(len(key) > 0)
        key_widget.remove_css_class('error')

        if not key:
            return
        
        counts = 0
        for kv_widgets in self.env_variables_widgets:
            k, v = kv_widgets
            
            if k.get_text() == key_widget.get_text():
                counts += 1

        if counts > 1:
            key_widget.add_css_class('error')
        
        value_widget.set_sensitive(counts == 1)
        self.save_vars_btn.set_sensitive(counts == 1)

    def on_save_env_vars_clicked(self, widget):
        widget.set_sensitive(False)
        self.update_env_variables()
        self.provider.update_desktop_file(self.app_list_element)

    def on_delete_env_var_clicked(self, widget, key_widget, value_widget, listbox):
        for i, kv_widgets in enumerate(self.env_variables_widgets):
            k, v = kv_widgets
            
            if k.get_text() == key_widget.get_text():
                self.env_variables_widgets.pop(i)
                break

        self.update_env_variables()
        self.provider.update_desktop_file(self.app_list_element)

        if self.env_variables_group_container:
            self.env_variables_group_container.remove(listbox)

    @debounce(0.5)
    @idle
    def on_cmd_arguments_changed(self, widget):
        text = widget.get_text().strip()
        text = text.replace('\n', '')
        widget.remove_css_class('error')

        self.app_list_element.exec_arguments = text

        try:
            self.provider.update_desktop_file(self.app_list_element)
        except:
            widget.add_css_class('error')

    # Returns the configuration from the json for this specific app
    def get_config_for_app(self) -> dict:
        return self.app_list_element.get_config()

    def on_web_browser_open_btn_clicked(self, widget):
        app_config = self.get_config_for_app()

        if ('website' in app_config) and url_is_valid(app_config['website']):
            launcher = Gtk.UriLauncher.new(app_config['website'])
            launcher.launch()

    def on_update_url_info_btn_clicked(self, widget):
        url = 'https://gearlever.mijorus.it/docs/overview'
        launcher = Gtk.UriLauncher.new(url)
        launcher.launch()

    @_async
    def on_refresh_metadata_btn_clicked(self, widget):
        self.show_row_spinner(True)
        GLib.idle_add(lambda: widget.set_sensitive(False))

        self.provider.reload_metadata(self.app_list_element)

        icon = self.provider.get_icon(self.app_list_element)
        self.provider.refresh_data(self.app_list_element)

        generation = self.provider.get_appimage_type(self.app_list_element)

        self.complete_load(icon, generation)

    def on_open_folder_clicked(self, widget):
        path = Gio.File.new_for_path(os.path.dirname(self.app_list_element.file_path))
        launcher = Gtk.FileLauncher.new(path)
        launcher.launch()

    def update_env_variables(self):
        self.app_list_element.env_variables = []
        for kv_widgets in self.env_variables_widgets:
            k, v = kv_widgets

            if k.has_css_class('error') or v.has_css_class('error'):
                continue

            key = k.get_text().strip()
            value = v.get_text().strip()
            value = shlex.quote(value)

            if key:
                self.app_list_element.env_variables.append(f'{key}={value}')

    def on_create_edit_row_btn_clicked(self, w):
        edit_form = self.create_edit_env_var_form()
        self.env_variables_group_container.append(edit_form)

    # Create widgets methods

    def create_edit_custom_website_row(self) -> Adw.EntryRow:
        app_config = self.get_config_for_app()

        title =  _('Add a website')
        text = ''

        if 'website' in app_config:
            title = _('Website')
            text = app_config['website']

        row = Adw.EntryRow(
            title=title,
            selectable=False,
            text=text
        )

        row_img = Gtk.Image(icon_name='gl-earth', pixel_size=self.ACTION_ROW_ICON_SIZE)
        row_btn = Gtk.Button(
            icon_name='gl-arrow2-top-right-symbolic', 
            valign=Gtk.Align.CENTER, 
            tooltip_text=_('Open URL'),
        )
        
        row_btn.connect('clicked', self.on_web_browser_open_btn_clicked)

        row.connect('changed', self.on_web_browser_input_apply)
        row.add_prefix(row_img)
        row.add_suffix(row_btn)

        return row

    def create_edit_update_url_row(self) -> Adw.EntryRow:
        app_config = Config.get_app_update_config(self.app_list_element)

        row_btn = Gtk.Button(
            icon_name='gl-info-symbolic', 
            valign=Gtk.Align.CENTER, 
            tooltip_text=_('How it works'),
        )

        row_btn.connect('clicked', self.on_update_url_info_btn_clicked)

        save_btn_content = Adw.ButtonContent(
            icon_name='gearlever-check-plain-symbolic',
            label=_(self.BTN_SAVE),
        )

        self.update_url_save_btn = Gtk.Button(child=save_btn_content, sensitive=True,
                                        css_classes=['suggested-action'])

        self.update_url_save_btn.connect('clicked', self.on_app_update_url_apply)

        btn_container = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=5, 
                                valign=Gtk.Align.CENTER)

        btn_container.append(row_btn)
        btn_container.append(self.update_url_save_btn)

        group = Adw.PreferencesGroup(
            title=_('Update management'),
            description=self.UPDATE_INFO_NOT_EMBEDDED,
            header_suffix=btn_container,
        )

        combo_model = Gtk.StringList()

        combo_model.append(_('Default'))

        selected_model_name = app_config.get('manager', None)
        selected_model: Optional[UpdateManager] = None

        self.update_url_source = Adw.ComboRow(
            title=_('Source'),
            subtitle=_('Select the source type'),
            model=combo_model
        )

        available_models = UpdateManagerChecker.get_models()

        for i, m in enumerate(available_models):
            combo_model.append(m.label)

            if selected_model_name == m.name:
                selected_model = m
                self.update_url_source.set_selected(i + 1)

        if not selected_model:
            self.update_url_source.set_selected(0)

        self.update_url_source.connect('notify::selected', lambda *args: self.set_update_information())
        self.update_url_form_rows = []

        group.add(self.update_url_source)
        self.set_update_information()

        return group

    def create_reload_metadata_row(self) -> Adw.EntryRow:
        row = Adw.ActionRow(selectable=False, activatable=True,
            title=(_('Reload metadata')), 
            subtitle=_('Update information like icon, version and description.\nUseful if the app updated itself.')
        )

        row_img = Gtk.Image(icon_name='gl-refresh', pixel_size=self.ACTION_ROW_ICON_SIZE)

        row.add_prefix(row_img)
        row.connect('activated', self.on_refresh_metadata_btn_clicked)

        return row

    def create_show_exec_args_row(self) -> Adw.ActionRow:
        row = Adw.EntryRow(
            title=(_('Command line arguments')),
            selectable=False,
            text=self.app_list_element.exec_arguments
        )

        row_img = Gtk.Image(icon_name='gearlever-cmd-args', pixel_size=self.ACTION_ROW_ICON_SIZE)
        row.connect('changed', self.on_cmd_arguments_changed)
        row.add_prefix(row_img)

        return row

    def create_app_hash_row(self) -> Adw.ActionRow:
        md5_hash = get_file_hash(Gio.File.new_for_path(self.app_list_element.file_path))
        sha1_hash = get_file_hash(Gio.File.new_for_path(self.app_list_element.file_path), 'sha1')
        sha256_hash = get_file_hash(Gio.File.new_for_path(self.app_list_element.file_path), 'sha256')

        row = Adw.ActionRow(
            subtitle=f'md5: {md5_hash}\nsha1: {sha1_hash}\nsha256: {sha256_hash}', 
            title=_('Hash'),
            selectable=True
        )

        row_img = Gtk.Image(icon_name='gl-hash-symbolic', pixel_size=self.ACTION_ROW_ICON_SIZE)
        row.add_prefix(row_img)

        return row
    
    def create_exec_path_row(self) -> Adw.ActionRow:
        row = Adw.ActionRow(title=_('Path'), subtitle=self.app_list_element.file_path, subtitle_selectable=True, selectable=False)
        row_img = Gtk.Image(icon_name='gearlever-file-manager-symbolic', pixel_size=self.ACTION_ROW_ICON_SIZE)
        row_btn = Gtk.Button(icon_name='gl-arrow2-top-right-symbolic', valign=Gtk.Align.CENTER, tooltip_text=_('Open Folder'))
        row_btn.connect('clicked', self.on_open_folder_clicked)
        row.add_prefix(row_img)
        row.add_suffix(row_btn)

        return row

    def create_package_info_row(self, gen) -> Adw.ActionRow:
        row = Adw.ActionRow(
            subtitle=f'{self.provider.name.capitalize()} Type. {gen}', 
            title=_('Package type'),
            selectable=False
        )

        row_img = Gtk.Image(resource=self.provider.icon, pixel_size=self.ACTION_ROW_ICON_SIZE)
        row.add_prefix(row_img)

        return row
    
    def create_edit_env_var_form(self, key='', value=''):
        listbox = Gtk.Box(
            orientation=Gtk.Orientation.HORIZONTAL,
            halign=Gtk.Align.CENTER,
            spacing=10,
            margin_top=self.EXTRA_DATA_SPACING / 2,
            margin_bottom=self.EXTRA_DATA_SPACING / 2,
        )

        row_key = Gtk.Entry(placeholder_text=_('Key'), text=key, hexpand=True)
        row_value = Gtk.Entry(placeholder_text=_('Value'), text=value, hexpand=True, sensitive=(len(key) > 0))
        delete_btn = Gtk.Button(icon_name='gl-user-trash-symbolic', css_classes=['destructive-action'])

        row_key.connect('changed', self.on_env_var_value_changed, row_key, row_value)
        row_value.connect('changed', self.on_env_var_value_changed, row_key, row_value)
        delete_btn.connect('clicked', self.on_delete_env_var_clicked, row_key, row_value, listbox)

        listbox.append(row_key)
        listbox.append(Gtk.Label.new('='))
        listbox.append(row_value)
        listbox.append(delete_btn)

        self.env_variables_widgets.append([row_key, row_value])

        return listbox
    
    def create_edit_env_vars_row(self) -> Adw.PreferencesGroup:
        add_btn_content = Adw.ButtonContent(
            icon_name='gl-plus-symbolic',
        )

        save_btn_content = Adw.ButtonContent(
            icon_name='gearlever-check-plain-symbolic',
            label=_(self.BTN_SAVE)
        )

        add_item_btn = Gtk.Button(child=add_btn_content)
        self.save_vars_btn = Gtk.Button(child=save_btn_content, sensitive=False,
                                        css_classes=['suggested-action'])


        self.save_vars_btn.connect('clicked', self.on_save_env_vars_clicked)

        btn_container = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=5, 
                                valign=Gtk.Align.CENTER)

        [btn_container.append(w) for w in [add_item_btn, self.save_vars_btn]]

        group = Adw.PreferencesGroup(
            title=_('Environment variables'),
            description=_('Add or customize environment for this application'),
            header_suffix=btn_container,
        )

        self.env_variables_group_container = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            css_classes=['card'],
            margin_top=self.EXTRA_DATA_SPACING / 2
        )

        group.add(self.env_variables_group_container)

        add_item_btn.connect('clicked', self.on_create_edit_row_btn_clicked)

        self.env_variables_widgets = []
        for kv in self.app_list_element.env_variables:
            k, v = kv.split('=', 1)

            row = self.create_edit_env_var_form(k, v)
            self.env_variables_group_container.append(row)

        return group



import threading

import gi

from gi.repository import GLib, GObject

# Thank you to https://github.com/linuxmint/webapp-manager
# check out common.py for the idea

# Used as a decorator to run things in the background
def _async(func):
    def wrapper(*args, **kwargs):
        thread = threading.Thread(target=func, args=args, kwargs=kwargs)
        thread.daemon = True
        thread.start()
        return thread
    return wrapper

def _async_keepalive(func):
    def wrapper(*args, **kwargs):
        thread = threading.Thread(target=func, args=args, kwargs=kwargs)
        thread.daemon = False
        thread.start()
        return thread
    return wrapper

# Used as a decorator to run things in the main loop, from another thread
def idle(func):
    def wrapper(*args, **kwargs):
        GLib.idle_add(func, *args)
    return wrapper


def debounce(wait_time):
    """
    Decorator that will debounce a function so that it is called after wait_time seconds
    If it is called multiple times, will wait for the last call to be debounced and run only this one.
    See the test_debounce.py file for examples
    """

    def decorator(function):
        def debounced(*args, **kwargs):
            def call_function():
                debounced._timer = None
                return function(*args, **kwargs)

            if debounced._timer is not None:
                debounced._timer.cancel()

            debounced._timer = threading.Timer(wait_time, call_function)
            debounced._timer.start()

        debounced._timer = None
        return debounced

    return decorator


import logging
import requests
import os
import re
from fnmatch import fnmatch
from typing import Optional, Literal
from gi.repository import Adw, Gio
from urllib.parse import urlsplit, urljoin

from ..lib import json_config
from ..lib.ini_config import Config
from .UpdateManager import UpdateManager
from .StaticFileUpdater import StaticFileUpdater
from ..components.AdwEntryRowDefault import AdwEntryRowDefault

class ForgejoUpdater(UpdateManager):
    # Example https://git.citron-emu.org/Citron/Emulator

    staticfile_manager: Optional[StaticFileUpdater]
    label = 'Forgejo'
    name = 'ForgejoUpdater'

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.staticfile_manager = None
        self.embedded = False
        self.repo_url_row = None
        self.repo_filename_row = None
        self.allow_prereleases_row = None

    def get_url_data(self, url: str):
        paths = []
        if url.startswith('https://'):
            logging.debug(f'ForgejoUpdater: found http url, trying to detect forgejo data')
            urldata = urlsplit(url)

            paths = urldata.path.split('/')

            if len(paths) < 3:
                return None

            return {
                'netloc': urldata.netloc,
                'username': paths[1],
                'repo': paths[2],
            }
        
        return None

    def migrate_v2(self):
        app_config = json_config.read_config_for_app(self.el)
        config = None

        if 'update_manager_config' in app_config:
            url_data = self.get_url_data(app_config['update_url'])
            urldata = urlsplit(app_config['update_url'])
            paths = urldata.path.split('/')

            if url_data and len(paths) >= 7:
                config = {
                    'allow_prereleases': app_config.get('update_manager_config', {}).get('allow_prereleases', False),
                    'repo_url': '/'.join(['https://', url_data['netloc'], url_data['username'], url_data['repo']]),
                    'repo_filename': paths[6],
                }

        if config:
            Config.set_app_update_config(self.el, self, config)

    def download(self, status_update_cb) -> tuple[str, str]:
        target_asset = self.fetch_target_asset()
        if not target_asset:
            raise Exception(f'Missing target_asset for {self.name} instance')

        dwnl = target_asset['browser_download_url']
        self.staticfile_manager = StaticFileUpdater(self.el, dwnl)
        fname, etag = self.staticfile_manager.download(status_update_cb)

        self.staticfile_manager = None
        return fname, target_asset['id']

    def cancel_download(self):
        if self.staticfile_manager:
            self.staticfile_manager.cancel_download()
            self.staticfile_manager = None

    def cleanup(self):
        if self.staticfile_manager:
            self.staticfile_manager.cleanup()

    def fetch_target_asset(self):
        conf = self.get_config()
        url_data = self.get_url_data(conf.get('repo_url', ''))

        if not url_data:
            return

        api_version = 'v1'
        allow_prereleases = conf.getboolean('allow_prereleases', False)

        rel_url = '/'.join([
            f'https://{url_data["netloc"]}',
            'api', 
            api_version,
            'repos',
            url_data["username"],
            url_data["repo"],
            'releases'
        ])

        if not allow_prereleases:
            rel_url += '/latest'

        try:
            rel_data_resp = requests.get(rel_url)
            rel_data_resp.raise_for_status()
            rel_data = rel_data_resp.json()
        except Exception as e:
            logging.error(e)
            return

        release = None

        if allow_prereleases:
            for r in rel_data:
                if r['draft'] == False:
                    release = r
                    break
        else:
            release = rel_data

        if not release:
            logging.error('Empty release list')
            return

        logging.debug(f'Found {len(release["assets"])} assets from {rel_url}')

        if not conf.get('repo_filename'):
            return

        download_asset = None
        possible_targets = []
        for asset in release['assets']:
            if fnmatch(asset['name'], conf.get('repo_filename', '')):
                possible_targets.append(asset)

        if len(possible_targets) == 1:
            download_asset = possible_targets[0]
        else:
            logging.info(f'found {len(possible_targets)} possible file targets')

            for t in possible_targets:
                logging.info(' - ' + t['name'])

        if not download_asset:
            logging.debug(f'No matching assets found from {rel_url}')
            return

        logging.debug(f'Found 1 matching asset: {download_asset["browser_download_url"]}')
        return download_asset

    def is_update_available(self):
        config = self.get_config()
        target_asset = self.fetch_target_asset()

        if target_asset:
            old_size = os.path.getsize(self.el.file_path)
            is_size_different = target_asset['size'] != old_size
            return is_size_different

        if config.get('repo_filename'):
            return None

        return False

    def load_form_rows(self, embedded=None): 
        config = self.get_config()
        repo_url = config.get('repo_url')
        filename = config.get('repo_filename')

        self.repo_url_row = AdwEntryRowDefault(
            text=repo_url,
            icon_name='gl-git',
            sensitive=True,
            title=_('Repo URL')
        )

        self.repo_filename_row = AdwEntryRowDefault(
            text=filename,
            icon_name='gl-paper',
            sensitive=True,
            title=_('Release file name')
        )

        self.allow_prereleases_row = Adw.SwitchRow(
            title=_('Allow pre-releases'),
            active=config.getboolean('allow_prereleases', False)
        )

        return [
            self.repo_url_row, 
            self.repo_filename_row,
            self.allow_prereleases_row
        ]

    def get_config_from_form(self):
        allow_prereleases = False
        repo_url = None
        repo_filename = None

        if self.allow_prereleases_row:
            allow_prereleases = self.allow_prereleases_row.get_active()

        if self.repo_url_row:
            repo_url = self.repo_url_row.get_text().strip()

        if self.repo_filename_row:
            repo_filename = self.repo_filename_row.get_text().strip()

        return {
            'allow_prereleases': allow_prereleases,
            'repo_url': repo_url,
            'repo_filename': repo_filename,
        }
    
    def validate_config(self, config):
        data = self.get_url_data(config['repo_url'])

        if not data:
            raise Exception(f'Invalid {self.name} url')



from gi.repository import Gtk, GObject, Adw, Gdk, Gio, Pango, GLib

class AdwEntryRowDefault(Adw.EntryRow):
    ACTION_ROW_ICON_SIZE = 34

    def __init__(self, default_text=None, icon_name=None, **kwargs):
        super().__init__(**kwargs)
        self.default_text = default_text
        self.row_btn = None

        if icon_name:
            row_img = Gtk.Image(icon_name=icon_name, pixel_size=self.ACTION_ROW_ICON_SIZE)
            self.add_prefix(row_img)

        self.connect('changed', self.on_changed)

        self.row_btn = Gtk.Button(
            icon_name='gl-undo', 
            valign=Gtk.Align.CENTER,
            visible=False
        )
        
        self.row_btn.connect('clicked', self.on_reset_default_clicked)
        self.row_btn.set_sensitive(self.default_text != self.get_text())

        self.add_suffix(self.row_btn)
        self.set_default_text(default_text)

    def on_reset_default_clicked(self, *args):
        self.set_text(self.default_text)

    def on_changed(self, *args):
        if self.row_btn:
            self.row_btn.set_sensitive(self.default_text != self.get_text())

    def set_default_text(self, text: str | None):
        self.default_text = text
        self.row_btn.set_visible(text != None)
        self.row_btn.set_sensitive(self.default_text != self.get_text())



import logging
import json
import gi
import os
import configparser
import base64

gi.require_version('Gtk', '4.0')
gi.require_version('Adw', '1')

from gi.repository import GLib  # noqa


def read_json_config(name: str):
    path = f'{GLib.get_user_config_dir()}/{name}.json'
    logging.debug(f'Reading config from {path}')

    if not os.path.isfile(path):
        return {}

    with open(path, 'r') as file:
        return json.loads(file.read() or '{}')

def read_config_for_app(el) -> dict:
    conf = read_json_config('apps')
    b64name = base64.b64encode(el.name.encode('utf8')).decode('ascii')

    app_config = conf[b64name] if b64name in conf else {}
    app_config['b64name'] = b64name

    return app_config
